@@ -2,43 +2,51 @@
 #include "ControlSocket.h"
 #include "directorycache.h"
 #include "engineprivate.h"
-#include "event_loop.h"
-#include "ftpcontrolsocket.h"
-#include "httpcontrolsocket.h"
+#include "ftp/ftpcontrolsocket.h"
+#include "http/httpcontrolsocket.h"
 #include "logging_private.h"
 #include "pathcache.h"
 #include "ratelimiter.h"
-#include "sftpcontrolsocket.h"
+#include "sftp/sftpcontrolsocket.h"
+
+#include <libfilezilla/event_loop.hpp>
 
 #include <algorithm>
 
-mutex CFileZillaEnginePrivate::mutex_;
+fz::mutex CFileZillaEnginePrivate::mutex_;
 std::vector<CFileZillaEnginePrivate*> CFileZillaEnginePrivate::m_engineList;
 std::atomic_int CFileZillaEnginePrivate::m_activeStatus[2] = {{0}, {0}};
 std::list<CFileZillaEnginePrivate::t_failedLogins> CFileZillaEnginePrivate::m_failedLogins;
 
-CFileZillaEnginePrivate::CFileZillaEnginePrivate(CFileZillaEngineContext& context, CFileZillaEngine& parent)
-	: CEventHandler(context.GetEventLoop())
+namespace {
+unsigned int get_next_engine_id(fz::mutex& mutex)
+{
+	fz::scoped_lock lock(mutex);
+	static unsigned int id = 0;
+	return ++id;
+}
+}
+
+CFileZillaEnginePrivate::CFileZillaEnginePrivate(CFileZillaEngineContext& context, CFileZillaEngine& parent, EngineNotificationHandler& notificationHandler)
+	: event_handler(context.GetEventLoop())
 	, transfer_status_(*this)
+	, notification_handler_(notificationHandler)
+	, m_engine_id(get_next_engine_id(mutex_))
 	, m_options(context.GetOptions())
 	, m_rateLimiter(context.GetRateLimiter())
 	, directory_cache_(context.GetDirectoryCache())
 	, path_cache_(context.GetPathCache())
 	, parent_(parent)
+	, thread_pool_(context.GetThreadPool())
+	, encoding_converter_(context.GetCustomEncodingConverter())
 {
 	m_engineList.push_back(this);
 
-	{
-		scoped_lock lock(mutex_);
-		static int id = 0;
-		m_engine_id = ++id;
-	}
-
 	m_pLogging = new CLogging(*this);
 
 	{
 		bool queue_logs = ShouldQueueLogsFromOptions();
-		scoped_lock lock(notification_mutex_);
+		fz::scoped_lock lock(notification_mutex_);
 		queue_logs_ = queue_logs;
 	}
 
@@ -57,7 +65,7 @@ bool CFileZillaEnginePrivate::ShouldQueueLogsFromOptions() const
 
 CFileZillaEnginePrivate::~CFileZillaEnginePrivate()
 {
-	RemoveHandler();
+	remove_handler();
 	m_maySendNotificationEvent = false;
 
 	m_pControlSocket.reset();
@@ -79,8 +87,9 @@ CFileZillaEnginePrivate::~CFileZillaEnginePrivate()
 
 	delete m_pLogging;
 
-	if (m_engineList.empty())
-		CSocket::Cleanup(true);
+	if (m_engineList.empty()) {
+		fz::socket::cleanup(true);
+	}
 }
 
 void CFileZillaEnginePrivate::OnEngineEvent(EngineNotificationType type)
@@ -90,9 +99,6 @@ void CFileZillaEnginePrivate::OnEngineEvent(EngineNotificationType type)
 	case engineCancel:
 		DoCancel();
 		break;
-	case engineTransferEnd:
-		if (m_pControlSocket)
-			m_pControlSocket->TransferEnd();
 	default:
 		break;
 	}
@@ -100,64 +106,71 @@ void CFileZillaEnginePrivate::OnEngineEvent(EngineNotificationType type)
 
 bool CFileZillaEnginePrivate::IsBusy() const
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	return m_pCurrentCommand != 0;
 }
 
 bool CFileZillaEnginePrivate::IsConnected() const
 {
-	scoped_lock lock(mutex_);
-	if (!m_pControlSocket)
+	fz::scoped_lock lock(mutex_);
+	if (!m_pControlSocket) {
 		return false;
+	}
 
 	return m_pControlSocket->Connected();
 }
 
 const CCommand *CFileZillaEnginePrivate::GetCurrentCommand() const
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	return m_pCurrentCommand.get();
 }
 
 Command CFileZillaEnginePrivate::GetCurrentCommandId() const
 {
-	scoped_lock lock(mutex_);
-	if (!m_pCurrentCommand)
+	fz::scoped_lock lock(mutex_);
+	if (!m_pCurrentCommand) {
 		return Command::none;
-	else
+	}
+	else {
 		return GetCurrentCommand()->GetId();
+	}
+}
+
+void CFileZillaEnginePrivate::AddNotification(fz::scoped_lock& lock, CNotification *pNotification)
+{
+	m_NotificationList.push_back(pNotification);
+
+	if (m_maySendNotificationEvent) {
+		m_maySendNotificationEvent = false;
+		lock.unlock();
+		notification_handler_.OnEngineEvent(&parent_);
+	}
 }
 
 void CFileZillaEnginePrivate::AddNotification(CNotification *pNotification)
 {
-	{
-		scoped_lock lock(notification_mutex_);
-		m_NotificationList.push_back(pNotification);
-
-		if (!m_maySendNotificationEvent || !m_pEventHandler) {
-			return;
-		}
-		m_maySendNotificationEvent = false;
-	}
-
-	m_pEventHandler->QueueEvent(new wxFzEvent(&parent_));
+	fz::scoped_lock lock(notification_mutex_);
+	AddNotification(lock, pNotification);
 }
 
 void CFileZillaEnginePrivate::AddLogNotification(CLogmsgNotification *pNotification)
 {
-	scoped_lock lock(notification_mutex_);
+	fz::scoped_lock lock(notification_mutex_);
 
 	if (pNotification->msgType == MessageType::Error) {
 		queue_logs_ = false;
-		SendQueuedLogs();
-		AddNotification(pNotification);
+
+		m_NotificationList.insert(m_NotificationList.end(), queued_logs_.begin(), queued_logs_.end());
+		queued_logs_.clear();
+		AddNotification(lock, pNotification);
 	}
 	else if (pNotification->msgType == MessageType::Status) {
-		ClearQueuedLogs(false);
-		AddNotification(pNotification);
+		ClearQueuedLogs(lock, false);
+		AddNotification(lock, pNotification);
 	}
 	else if (!queue_logs_) {
-		AddNotification(pNotification);
+		AddNotification(lock, pNotification);
 	}
 	else {
 		queued_logs_.push_back(pNotification);
@@ -167,7 +180,7 @@ void CFileZillaEnginePrivate::AddLogNotification(CLogmsgNotification *pNotificat
 void CFileZillaEnginePrivate::SendQueuedLogs(bool reset_flag)
 {
 	{
-		scoped_lock lock(notification_mutex_);
+		fz::scoped_lock lock(notification_mutex_);
 		m_NotificationList.insert(m_NotificationList.end(), queued_logs_.begin(), queued_logs_.end());
 		queued_logs_.clear();
 
@@ -175,19 +188,17 @@ void CFileZillaEnginePrivate::SendQueuedLogs(bool reset_flag)
 			queue_logs_ = ShouldQueueLogsFromOptions();
 		}
 
-		if (!m_maySendNotificationEvent || !m_pEventHandler || m_NotificationList.empty()) {
+		if (!m_maySendNotificationEvent || m_NotificationList.empty()) {
 			return;
 		}
 		m_maySendNotificationEvent = false;
 	}
 
-	m_pEventHandler->QueueEvent(new wxFzEvent(&parent_));
+	notification_handler_.OnEngineEvent(&parent_);
 }
 
-void CFileZillaEnginePrivate::ClearQueuedLogs(bool reset_flag)
+void CFileZillaEnginePrivate::ClearQueuedLogs(fz::scoped_lock&, bool reset_flag)
 {
-	scoped_lock lock(notification_mutex_);
-
 	for (auto msg : queued_logs_) {
 		delete msg;
 	}
@@ -198,13 +209,16 @@ void CFileZillaEnginePrivate::ClearQueuedLogs(bool reset_flag)
 	}
 }
 
+void CFileZillaEnginePrivate::ClearQueuedLogs(bool reset_flag)
+{
+	fz::scoped_lock lock(notification_mutex_);
+	ClearQueuedLogs(lock, reset_flag);
+}
+
 int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 {
-	scoped_lock lock(mutex_);
-	m_pLogging->LogMessage(MessageType::Debug_Debug, _T("CFileZillaEnginePrivate::ResetOperation(%d)"), nErrorCode);
-
-	if (nErrorCode & FZ_REPLY_DISCONNECTED)
-		m_lastListDir.clear();
+	fz::scoped_lock lock(mutex_);
+	m_pLogging->LogMessage(MessageType::Debug_Debug, L"CFileZillaEnginePrivate::ResetOperation(%d)", nErrorCode);
 
 	if (m_pCurrentCommand) {
 		if ((nErrorCode & FZ_REPLY_NOTSUPPORTED) == FZ_REPLY_NOTSUPPORTED) {
@@ -222,12 +236,13 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 				if ((nErrorCode & FZ_REPLY_CRITICALERROR) != FZ_REPLY_CRITICALERROR) {
 					++m_retryCount;
 					if (m_retryCount < m_options.GetOptionVal(OPTION_RECONNECTCOUNT) && connectCommand.RetryConnecting()) {
-						unsigned int delay = GetRemainingReconnectDelay(connectCommand.GetServer());
-						if (!delay)
-							delay = 1;
+						fz::duration delay = GetRemainingReconnectDelay(connectCommand.GetServer());
+						if (!delay) {
+							delay = fz::duration::from_seconds(1);
+						}
 						m_pLogging->LogMessage(MessageType::Status, _("Waiting to retry..."));
-						StopTimer(m_retryTimer);
-						m_retryTimer = AddTimer(duration::from_milliseconds(delay), true);
+						stop_timer(m_retryTimer);
+						m_retryTimer = add_timer(delay, true);
 						return FZ_REPLY_WOULDBLOCK;
 					}
 				}
@@ -240,8 +255,9 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 			notification->commandId = m_pCurrentCommand->GetId();
 			AddNotification(notification);
 		}
-		else
+		else {
 			m_nControlSocketError |= nErrorCode;
+		}
 
 		m_pCurrentCommand.reset();
 	}
@@ -266,15 +282,16 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 
 unsigned int CFileZillaEnginePrivate::GetNextAsyncRequestNumber()
 {
-	scoped_lock lock(notification_mutex_);
+	fz::scoped_lock lock(notification_mutex_);
 	return ++m_asyncRequestCounter;
 }
 
 // Command handlers
-int CFileZillaEnginePrivate::Connect(const CConnectCommand &command)
+int CFileZillaEnginePrivate::Connect(CConnectCommand const& command)
 {
-	if (IsConnected())
+	if (IsConnected()) {
 		return FZ_REPLY_ALREADYCONNECTED;
+	}
 
 	m_retryCount = 0;
 
@@ -286,8 +303,9 @@ int CFileZillaEnginePrivate::Connect(const CConnectCommand &command)
 
 	if (command.GetServer().GetPort() != CServer::GetDefaultPort(command.GetServer().GetProtocol())) {
 		ServerProtocol protocol = CServer::GetProtocolFromPort(command.GetServer().GetPort(), true);
-		if (protocol != UNKNOWN && protocol != command.GetServer().GetProtocol())
+		if (protocol != UNKNOWN && protocol != command.GetServer().GetProtocol()) {
 			m_pLogging->LogMessage(MessageType::Status, _("Selected port usually in use by a different protocol."));
+		}
 	}
 
 	return ContinueConnect();
@@ -304,29 +322,29 @@ int CFileZillaEnginePrivate::Disconnect(CDisconnectCommand const&)
 	return res;
 }
 
-int CFileZillaEnginePrivate::List(const CListCommand &command)
+int CFileZillaEnginePrivate::List(CListCommand const& command)
 {
 	int flags = command.GetFlags();
 	bool const refresh = (command.GetFlags() & LIST_FLAG_REFRESH) != 0;
 	bool const avoid = (command.GetFlags() & LIST_FLAG_AVOID) != 0;
 
 	if (!refresh && !command.GetPath().empty()) {
-		const CServer* pServer = m_pControlSocket->GetCurrentServer();
-		if (pServer) {
-			CServerPath path(path_cache_.Lookup(*pServer, command.GetPath(), command.GetSubDir()));
-			if (path.empty() && command.GetSubDir().empty())
+		CServer const& server =m_pControlSocket->GetCurrentServer();
+		if (server) {
+			CServerPath path(path_cache_.Lookup(server, command.GetPath(), command.GetSubDir()));
+			if (path.empty() && command.GetSubDir().empty()) {
 				path = command.GetPath();
+			}
 			if (!path.empty()) {
 				CDirectoryListing *pListing = new CDirectoryListing;
 				bool is_outdated = false;
-				bool found = directory_cache_.Lookup(*pListing, *pServer, path, true, is_outdated);
+				bool found = directory_cache_.Lookup(*pListing, server, path, true, is_outdated);
 				if (found && !is_outdated) {
-					if (pListing->get_unsure_flags())
+					if (pListing->get_unsure_flags()) {
 						flags |= LIST_FLAG_REFRESH;
+					}
 					else {
 						if (!avoid) {
-							m_lastListDir = pListing->path;
-							m_lastListTime = CMonotonicClock::now();
 							CDirectoryListingNotification *pNotification = new CDirectoryListingNotification(pListing->path);
 							AddNotification(pNotification);
 						}
@@ -334,28 +352,32 @@ int CFileZillaEnginePrivate::List(const CListCommand &command)
 						return FZ_REPLY_OK;
 					}
 				}
-				if (is_outdated)
+				if (is_outdated) {
 					flags |= LIST_FLAG_REFRESH;
+				}
 				delete pListing;
 			}
 		}
 	}
 
-	return m_pControlSocket->List(command.GetPath(), command.GetSubDir(), flags);
+	m_pControlSocket->List(command.GetPath(), command.GetSubDir(), flags);
+	return FZ_REPLY_CONTINUE;
 }
 
-int CFileZillaEnginePrivate::FileTransfer(const CFileTransferCommand &command)
+int CFileZillaEnginePrivate::FileTransfer(CFileTransferCommand const& command)
 {
-	return m_pControlSocket->FileTransfer(command.GetLocalFile(), command.GetRemotePath(), command.GetRemoteFile(), command.Download(), command.GetTransferSettings());
+	m_pControlSocket->FileTransfer(command.GetLocalFile(), command.GetRemotePath(), command.GetRemoteFile(), command.Download(), command.GetTransferSettings());
+	return FZ_REPLY_CONTINUE;
 }
 
-int CFileZillaEnginePrivate::RawCommand(const CRawCommand& command)
+int CFileZillaEnginePrivate::RawCommand(CRawCommand const& command)
 {
 	{
-		scoped_lock lock(notification_mutex_);
+		fz::scoped_lock lock(notification_mutex_);
 		queue_logs_ = false;
 	}
-	return m_pControlSocket->RawCommand(command.GetCommand());
+	m_pControlSocket->RawCommand(command.GetCommand());
+	return FZ_REPLY_CONTINUE;
 }
 
 int CFileZillaEnginePrivate::Delete(CDeleteCommand& command)
@@ -366,86 +388,40 @@ int CFileZillaEnginePrivate::Delete(CDeleteCommand& command)
 	else {
 		m_pLogging->LogMessage(MessageType::Status, _("Deleting %u files from \"%s\""), static_cast<unsigned int>(command.GetFiles().size()), command.GetPath().GetPath());
 	}
-	return m_pControlSocket->Delete(command.GetPath(), command.ExtractFiles());
+	m_pControlSocket->Delete(command.GetPath(), command.ExtractFiles());
+	return FZ_REPLY_CONTINUE;
 }
 
-int CFileZillaEnginePrivate::RemoveDir(const CRemoveDirCommand& command)
+int CFileZillaEnginePrivate::RemoveDir(CRemoveDirCommand const& command)
 {
-	return m_pControlSocket->RemoveDir(command.GetPath(), command.GetSubDir());
+	m_pControlSocket->RemoveDir(command.GetPath(), command.GetSubDir());
+	return FZ_REPLY_CONTINUE;
 }
 
-int CFileZillaEnginePrivate::Mkdir(const CMkdirCommand& command)
+int CFileZillaEnginePrivate::Mkdir(CMkdirCommand const& command)
 {
-	return m_pControlSocket->Mkdir(command.GetPath());
+	m_pControlSocket->Mkdir(command.GetPath());
+	return FZ_REPLY_CONTINUE;
 }
 
-int CFileZillaEnginePrivate::Rename(const CRenameCommand& command)
+int CFileZillaEnginePrivate::Rename(CRenameCommand const& command)
 {
-	return m_pControlSocket->Rename(command);
+	m_pControlSocket->Rename(command);
+	return FZ_REPLY_CONTINUE;
 }
 
-int CFileZillaEnginePrivate::Chmod(const CChmodCommand& command)
+int CFileZillaEnginePrivate::Chmod(CChmodCommand const& command)
 {
-	return m_pControlSocket->Chmod(command);
-}
-
-void CFileZillaEnginePrivate::SendDirectoryListingNotification(const CServerPath& path, bool onList, bool modified, bool failed)
-{
-	scoped_lock lock(mutex_);
-
-	wxASSERT(m_pControlSocket);
-
-	const CServer* const pOwnServer = m_pControlSocket->GetCurrentServer();
-	wxASSERT(pOwnServer);
-
-	m_lastListDir = path;
-
-	if (failed) {
-		AddNotification(new CDirectoryListingNotification(path, false, true));
-		m_lastListTime = CMonotonicClock::now();
-
-		// On failed messages, we don't notify other engines
-		return;
-	}
-
-	CMonotonicClock changeTime;
-	if (!directory_cache_.GetChangeTime(changeTime, *pOwnServer, path))
-		return;
-
-	AddNotification(new CDirectoryListingNotification(path, !onList));
-	m_lastListTime = changeTime;
-
-	if (!modified)
-		return;
-
-	// Iterate over the other engine, send notification if last listing
-	// directory is the same
-	for (auto & engine : m_engineList) {
-		if (!engine->m_pControlSocket || engine->m_pControlSocket == m_pControlSocket) {
-			continue;
-		}
-
-		CServer const* const pServer = engine->m_pControlSocket->GetCurrentServer();
-		if (!pServer || *pServer != *pOwnServer)
-			continue;
-
-		if (engine->m_lastListDir != path)
-			continue;
-
-		if (engine->m_lastListTime && changeTime <= engine->m_lastListTime)
-			continue;
-
-		engine->m_lastListTime = changeTime;
-		engine->AddNotification(new CDirectoryListingNotification(path, true));
-	}
+	m_pControlSocket->Chmod(command);
+	return FZ_REPLY_CONTINUE;
 }
 
 void CFileZillaEnginePrivate::RegisterFailedLoginAttempt(const CServer& server, bool critical)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	std::list<t_failedLogins>::iterator iter = m_failedLogins.begin();
 	while (iter != m_failedLogins.end()) {
-		duration const span = CDateTime::Now() - iter->time;
+		fz::duration const span = fz::monotonic_clock::now() - iter->time;
 		if (span.get_seconds() >= m_options.GetOptionVal(OPTION_RECONNECTDELAY) ||
 			iter->server == server || (!critical && (iter->server.GetHost() == server.GetHost() && iter->server.GetPort() == server.GetPort())))
 		{
@@ -453,41 +429,45 @@ void CFileZillaEnginePrivate::RegisterFailedLoginAttempt(const CServer& server,
 			++iter;
 			m_failedLogins.erase(prev);
 		}
-		else
+		else {
 			++iter;
+		}
 	}
 
 	t_failedLogins failure;
 	failure.server = server;
-	failure.time = CDateTime::Now();
+	failure.time = fz::monotonic_clock::now();
 	failure.critical = critical;
 	m_failedLogins.push_back(failure);
 }
 
-unsigned int CFileZillaEnginePrivate::GetRemainingReconnectDelay(const CServer& server)
+fz::duration CFileZillaEnginePrivate::GetRemainingReconnectDelay(CServer const& server)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	std::list<t_failedLogins>::iterator iter = m_failedLogins.begin();
 	while (iter != m_failedLogins.end()) {
-		duration const span = CDateTime::Now() - iter->time;
-		const int delay = m_options.GetOptionVal(OPTION_RECONNECTDELAY);
-		if (span.get_seconds() >= delay) {
+		fz::duration const span = fz::monotonic_clock::now() - iter->time;
+		fz::duration const delay = fz::duration::from_seconds(m_options.GetOptionVal(OPTION_RECONNECTDELAY));
+		if (span >= delay) {
 			std::list<t_failedLogins>::iterator prev = iter;
 			++iter;
 			m_failedLogins.erase(prev);
 		}
-		else if (!iter->critical && iter->server.GetHost() == server.GetHost() && iter->server.GetPort() == server.GetPort())
-			return delay * 1000 - span.get_milliseconds();
-		else if (iter->server == server)
-			return delay * 1000 - span.get_milliseconds();
-		else
+		else if (!iter->critical && iter->server.GetHost() == server.GetHost() && iter->server.GetPort() == server.GetPort()) {
+			return delay - span;
+		}
+		else if (iter->server == server) {
+			return delay - span;
+		}
+		else {
 			++iter;
+		}
 	}
 
-	return 0;
+	return fz::duration();
 }
 
-void CFileZillaEnginePrivate::OnTimer(int)
+void CFileZillaEnginePrivate::OnTimer(fz::timer_id)
 {
 	if (!m_retryTimer) {
 		return;
@@ -495,32 +475,41 @@ void CFileZillaEnginePrivate::OnTimer(int)
 	m_retryTimer = 0;
 
 	if (!m_pCurrentCommand || m_pCurrentCommand->GetId() != Command::connect) {
-		wxFAIL_MSG(_T("CFileZillaEnginePrivate::OnTimer called without pending Command::connect"));
+		m_pLogging->LogMessage(MessageType::Debug_Warning, L"CFileZillaEnginePrivate::OnTimer called without pending Command::connect");
 		return;
 	}
-	wxASSERT(!IsConnected());
 
 	m_pControlSocket.reset();
 
-	ContinueConnect();
+	int res = ContinueConnect();
+	res |= m_nControlSocketError;
+	m_nControlSocketError = 0;
+
+	if (res == FZ_REPLY_CONTINUE) {
+		assert(m_pControlSocket);
+		m_pControlSocket->SendNextCommand();
+	}
+	else if (res != FZ_REPLY_WOULDBLOCK) {
+		ResetOperation(res);
+	}
 }
 
 int CFileZillaEnginePrivate::ContinueConnect()
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 
 	if (!m_pCurrentCommand || m_pCurrentCommand->GetId() != Command::connect) {
-		m_pLogging->LogMessage(MessageType::Debug_Warning, _T("CFileZillaEnginePrivate::ContinueConnect called without pending Command::connect"));
+		m_pLogging->LogMessage(MessageType::Debug_Warning, L"CFileZillaEnginePrivate::ContinueConnect called without pending Command::connect");
 		return ResetOperation(FZ_REPLY_INTERNALERROR);
 	}
 
 	const CConnectCommand *pConnectCommand = static_cast<CConnectCommand *>(m_pCurrentCommand.get());
 	const CServer& server = pConnectCommand->GetServer();
-	unsigned int delay = GetRemainingReconnectDelay(server);
+	fz::duration const& delay = GetRemainingReconnectDelay(server);
 	if (delay) {
-		m_pLogging->LogMessage(MessageType::Status, wxPLURAL("Delaying connection for %d second due to previously failed connection attempt...", "Delaying connection for %d seconds due to previously failed connection attempt...", (delay + 999) / 1000), (delay + 999) / 1000);
-		StopTimer(m_retryTimer);
-		m_retryTimer = AddTimer(duration::from_milliseconds(delay), true);
+		m_pLogging->LogMessage(MessageType::Status, fztranslate("Delaying connection for %d second due to previously failed connection attempt...", "Delaying connection for %d seconds due to previously failed connection attempt...", (delay.get_milliseconds() + 999) / 1000), (delay.get_milliseconds() + 999) / 1000);
+		stop_timer(m_retryTimer);
+		m_retryTimer = add_timer(delay, true);
 		return FZ_REPLY_WOULDBLOCK;
 	}
 
@@ -540,45 +529,45 @@ int CFileZillaEnginePrivate::ContinueConnect()
 		m_pControlSocket = std::make_unique<CHttpControlSocket>(*this);
 		break;
 	default:
-		m_pLogging->LogMessage(MessageType::Debug_Warning, _T("Not a valid protocol: %d"), server.GetProtocol());
+		m_pLogging->LogMessage(MessageType::Debug_Warning, L"Not a valid protocol: %d", server.GetProtocol());
 		return FZ_REPLY_SYNTAXERROR|FZ_REPLY_DISCONNECTED;
 	}
 
-	int res = m_pControlSocket->Connect(server);
-	if (m_retryTimer)
-		return FZ_REPLY_WOULDBLOCK;
-
-	return res;
+	m_pControlSocket->Connect(server);
+	return FZ_REPLY_CONTINUE;
 }
 
 void CFileZillaEnginePrivate::InvalidateCurrentWorkingDirs(const CServerPath& path)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 
-	wxASSERT(m_pControlSocket);
-	const CServer* const pOwnServer = m_pControlSocket->GetCurrentServer();
-	wxASSERT(pOwnServer);
+	assert(m_pControlSocket);
+	CServer const& ownServer  = m_pControlSocket->GetCurrentServer();
+	assert(ownServer);
 
 	for (auto & engine : m_engineList) {
-		if (engine == this)
+		if (!engine || engine == this) {
 			continue;
+		}
 
-		if (!engine->m_pControlSocket)
+		if (!engine->m_pControlSocket) {
 			continue;
+		}
 
-		const CServer* const pServer = engine->m_pControlSocket->GetCurrentServer();
-		if (!pServer || *pServer != *pOwnServer)
+		CServer const& server = engine->m_pControlSocket->GetCurrentServer();
+		if (server != ownServer) {
 			continue;
+		}
 
 		engine->m_pControlSocket->InvalidateCurrentWorkingDir(path);
 	}
 }
 
-void CFileZillaEnginePrivate::operator()(CEventBase const& ev)
+void CFileZillaEnginePrivate::operator()(fz::event_base const& ev)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 
-	Dispatch<CFileZillaEngineEvent, CCommandEvent, CAsyncRequestReplyEvent, CTimerEvent>(ev, this,
+	fz::dispatch<CFileZillaEngineEvent, CCommandEvent, CAsyncRequestReplyEvent, fz::timer_event>(ev, this,
 		&CFileZillaEnginePrivate::OnEngineEvent,
 		&CFileZillaEnginePrivate::OnCommandEvent,
 		&CFileZillaEnginePrivate::OnSetAsyncRequestReplyEvent,
@@ -605,7 +594,7 @@ int CFileZillaEnginePrivate::CheckCommandPreconditions(CCommand const& command,
 
 void CFileZillaEnginePrivate::OnCommandEvent()
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 
 	if (m_pCurrentCommand) {
 		CCommand & command = *m_pCurrentCommand;
@@ -650,31 +639,39 @@ void CFileZillaEnginePrivate::OnCommandEvent()
 			}
 		}
 
-		if (id != Command::disconnect)
+		if (id != Command::disconnect) {
 			res |= m_nControlSocketError;
-		else if (res & FZ_REPLY_DISCONNECTED)
+		}
+		else if (res & FZ_REPLY_DISCONNECTED) {
 			res = FZ_REPLY_OK;
+		}
 		m_nControlSocketError = 0;
 
-		if (res != FZ_REPLY_WOULDBLOCK)
+		if (res == FZ_REPLY_CONTINUE) {
+			assert(m_pControlSocket);
+			m_pControlSocket->SendNextCommand();
+		}
+		else if (res != FZ_REPLY_WOULDBLOCK) {
 			ResetOperation(res);
+		}
 	}
 }
 
 void CFileZillaEnginePrivate::DoCancel()
 {
-	scoped_lock lock(mutex_);
-	if (!IsBusy())
+	fz::scoped_lock lock(mutex_);
+	if (!IsBusy()) {
 		return;
+	}
 
 	if (m_retryTimer) {
-		wxASSERT(m_pCurrentCommand && m_pCurrentCommand->GetId() == Command::connect);
+		assert(m_pCurrentCommand && m_pCurrentCommand->GetId() == Command::connect);
 
 		m_pControlSocket.reset();
 
 		m_pCurrentCommand.reset();
 
-		StopTimer(m_retryTimer);
+		stop_timer(m_retryTimer);
 		m_retryTimer = 0;
 
 		m_pLogging->LogMessage(MessageType::Error, _("Connection attempt interrupted by user"));
@@ -686,23 +683,27 @@ void CFileZillaEnginePrivate::DoCancel()
 		ClearQueuedLogs(true);
 	}
 	else {
-		if (m_pControlSocket)
+		if (m_pControlSocket) {
 			m_pControlSocket->Cancel();
-		else
+		}
+		else {
 			ResetOperation(FZ_REPLY_CANCELED);
+		}
 	}
 }
 
 bool CFileZillaEnginePrivate::CheckAsyncRequestReplyPreconditions(std::unique_ptr<CAsyncRequestNotification> const& reply)
 {
-	if (!reply)
+	if (!reply) {
 		return false;
-	if (!IsBusy())
+	}
+	if (!IsBusy()) {
 		return false;
+	}
 
 	bool match;
 	{
-		scoped_lock l(notification_mutex_);
+		fz::scoped_lock l(notification_mutex_);
 		match = reply->requestNumber == m_asyncRequestCounter;
 	}
 
@@ -711,7 +712,7 @@ bool CFileZillaEnginePrivate::CheckAsyncRequestReplyPreconditions(std::unique_pt
 
 void CFileZillaEnginePrivate::OnSetAsyncRequestReplyEvent(std::unique_ptr<CAsyncRequestNotification> const& reply)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	if (!CheckAsyncRequestReplyPreconditions(reply)) {
 		return;
 	}
@@ -720,16 +721,9 @@ void CFileZillaEnginePrivate::OnSetAsyncRequestReplyEvent(std::unique_ptr<CAsync
 	m_pControlSocket->SetAsyncRequestReply(reply.get());
 }
 
-int CFileZillaEnginePrivate::Init(wxEvtHandler *pEventHandler)
-{
-	scoped_lock lock(mutex_);
-	m_pEventHandler = pEventHandler;
-	return FZ_REPLY_OK;
-}
-
 int CFileZillaEnginePrivate::Execute(const CCommand &command)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 
 	int res = CheckCommandPreconditions(command, true);
 	if (res != FZ_REPLY_OK) {
@@ -737,14 +731,14 @@ int CFileZillaEnginePrivate::Execute(const CCommand &command)
 	}
 
 	m_pCurrentCommand.reset(command.Clone());
-	SendEvent<CCommandEvent>();
+	send_event<CCommandEvent>();
 
 	return FZ_REPLY_WOULDBLOCK;
 }
 
 std::unique_ptr<CNotification> CFileZillaEnginePrivate::GetNextNotification()
 {
-	scoped_lock lock(notification_mutex_);
+	fz::scoped_lock lock(notification_mutex_);
 
 	if (m_NotificationList.empty()) {
 		m_maySendNotificationEvent = true;
@@ -758,25 +752,27 @@ std::unique_ptr<CNotification> CFileZillaEnginePrivate::GetNextNotification()
 
 bool CFileZillaEnginePrivate::SetAsyncRequestReply(std::unique_ptr<CAsyncRequestNotification> && pNotification)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	if (!CheckAsyncRequestReplyPreconditions(pNotification)) {
 		return false;
 	}
 
-	SendEvent<CAsyncRequestReplyEvent>(std::move(pNotification));
+	send_event<CAsyncRequestReplyEvent>(std::move(pNotification));
 
 	return true;
 }
 
 bool CFileZillaEnginePrivate::IsPendingAsyncRequestReply(std::unique_ptr<CAsyncRequestNotification> const& pNotification)
 {
-	if (!pNotification)
+	if (!pNotification) {
 		return false;
+	}
 
-	if (!IsBusy())
+	if (!IsBusy()) {
 		return false;
+	}
 
-	scoped_lock lock(notification_mutex_);
+	fz::scoped_lock lock(notification_mutex_);
 	return pNotification->requestNumber == m_asyncRequestCounter;
 }
 
@@ -808,38 +804,42 @@ CTransferStatus CFileZillaEnginePrivate::GetTransferStatus(bool &changed)
 int CFileZillaEnginePrivate::CacheLookup(const CServerPath& path, CDirectoryListing& listing)
 {
 	// TODO: Possible optimization: Atomically get current server. The cache has its own mutex.
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 
-	if (!IsConnected())
+	if (!IsConnected()) {
 		return FZ_REPLY_ERROR;
+	}
 
-	wxASSERT(m_pControlSocket->GetCurrentServer());
+	assert(m_pControlSocket->GetCurrentServer());
 
 	bool is_outdated = false;
-	if (!directory_cache_.Lookup(listing, *m_pControlSocket->GetCurrentServer(), path, true, is_outdated))
+	if (!directory_cache_.Lookup(listing, m_pControlSocket->GetCurrentServer(), path, true, is_outdated)) {
 		return FZ_REPLY_ERROR;
+	}
 
 	return FZ_REPLY_OK;
 }
 
 int CFileZillaEnginePrivate::Cancel()
 {
-	scoped_lock lock(mutex_);
-	if (!IsBusy())
+	fz::scoped_lock lock(mutex_);
+	if (!IsBusy()) {
 		return FZ_REPLY_OK;
+	}
 
-	SendEvent<CFileZillaEngineEvent>(engineCancel);
+	send_event<CFileZillaEngineEvent>(engineCancel);
 	return FZ_REPLY_WOULDBLOCK;
 }
 
 void CFileZillaEnginePrivate::OnOptionsChanged(changed_options_t const&)
 {
 	bool queue_logs = ShouldQueueLogsFromOptions();
-	scoped_lock lock(notification_mutex_);
-	queue_logs_ = queue_logs;
-
-	if (!queue_logs_) {
-		SendQueuedLogs();
+	if (queue_logs) {
+		fz::scoped_lock lock(notification_mutex_);
+		queue_logs_ = true;
+	}
+	else {
+		SendQueuedLogs(true);
 	}
 }
 
@@ -852,7 +852,7 @@ CTransferStatusManager::CTransferStatusManager(CFileZillaEnginePrivate& engine)
 void CTransferStatusManager::Reset()
 {
 	{
-		scoped_lock lock(mutex_);
+		fz::scoped_lock lock(mutex_);
 		status_.clear();
 		send_state_ = 0;
 	}
@@ -860,9 +860,9 @@ void CTransferStatusManager::Reset()
 	engine_.AddNotification(new CTransferStatusNotification());
 }
 
-void CTransferStatusManager::Init(wxFileOffset totalSize, wxFileOffset startOffset, bool list)
+void CTransferStatusManager::Init(int64_t totalSize, int64_t startOffset, bool list)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	if (startOffset < 0)
 		startOffset = 0;
 
@@ -872,30 +872,30 @@ void CTransferStatusManager::Init(wxFileOffset totalSize, wxFileOffset startOffs
 
 void CTransferStatusManager::SetStartTime()
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	if (!status_)
 		return;
 
-	status_.started = CDateTime::Now();
+	status_.started = fz::datetime::now();
 }
 
 void CTransferStatusManager::SetMadeProgress()
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	if (!status_)
 		return;
 
 	status_.madeProgress = true;
 }
 
-void CTransferStatusManager::Update(wxFileOffset transferredBytes)
+void CTransferStatusManager::Update(int64_t transferredBytes)
 {
 	CNotification* notification = 0;
 
 	{
 		int64_t oldOffset = currentOffset_.fetch_add(transferredBytes);
 		if (!oldOffset) {
-			scoped_lock lock(mutex_);
+			fz::scoped_lock lock(mutex_);
 			if (!status_) {
 				return;
 			}
@@ -915,7 +915,7 @@ void CTransferStatusManager::Update(wxFileOffset transferredBytes)
 
 CTransferStatus CTransferStatusManager::Get(bool &changed)
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	if (!status_) {
 		changed = false;
 		send_state_ = 0;
@@ -936,6 +936,6 @@ CTransferStatus CTransferStatusManager::Get(bool &changed)
 
 bool CTransferStatusManager::empty()
 {
-	scoped_lock lock(mutex_);
+	fz::scoped_lock lock(mutex_);
 	return status_.empty();
 }
