@@ -7,39 +7,15 @@
 #include "Options.h"
 #include "wrapengine.h"
 #include "buildinfo.h"
-#include <wx/tokenzr.h>
 #include "cmdline.h"
 #include "welcome_dialog.h"
-#include <msgbox.h>
-#include "local_filesys.h"
+#include "msgbox.h"
 
-#include <wx/xrc/xh_animatctrl.h>
-#include <wx/xrc/xh_bmpbt.h>
-#include <wx/xrc/xh_bttn.h>
-#include <wx/xrc/xh_chckb.h>
-#include <wx/xrc/xh_chckl.h>
-#include <wx/xrc/xh_choic.h>
-#include <wx/xrc/xh_dlg.h>
-#include <wx/xrc/xh_gauge.h>
-#include <wx/xrc/xh_listb.h>
-#include <wx/xrc/xh_listc.h>
-#include <wx/xrc/xh_menu.h>
-#include <wx/xrc/xh_notbk.h>
-#include <wx/xrc/xh_panel.h>
-#include <wx/xrc/xh_radbt.h>
-#include <wx/xrc/xh_scwin.h>
-#include <wx/xrc/xh_sizer.h>
-#include <wx/xrc/xh_spin.h>
-#include <wx/xrc/xh_stbmp.h>
-#include <wx/xrc/xh_stbox.h>
-#include <wx/xrc/xh_stlin.h>
-#include <wx/xrc/xh_sttxt.h>
-#include "xh_text_ex.h"
-#include <wx/xrc/xh_tree.h>
-#include <wx/xrc/xh_hyperlink.h>
-#include "xh_toolb_ex.h"
+#include <libfilezilla/local_filesys.hpp>
+
+#include "wxfilesystem_blob_handler.h"
+#include "xrc_helper.h"
 #ifdef __WXMSW__
-#include <wx/socket.h>
 #include <wx/dynlib.h>
 #endif
 #ifdef WITH_LIBDBUS
@@ -52,6 +28,10 @@
 
 #include "locale_initializer.h"
 
+#ifdef USE_MAC_SANDBOX
+#include "osx_sandbox_userdirs.h"
+#endif
+
 #ifdef ENABLE_BINRELOC
 	#define BR_PTHREADS 0
 	#include "prefix.h"
@@ -63,10 +43,14 @@ IMPLEMENT_APP(CFileZillaApp)
 IMPLEMENT_APP_NO_MAIN(CFileZillaApp)
 #endif //__WXGTK__
 
+#if !wxUSE_PRINTF_POS_PARAMS
+  #error Please build wxWidgets with support for positional arguments.
+#endif
+
 CFileZillaApp::CFileZillaApp()
 {
-	m_profilingActive = true;
-	AddStartupProfileRecord(_T("CFileZillaApp::CFileZillaApp()"));
+	m_profile_start = fz::monotonic_clock::now();
+	AddStartupProfileRecord("CFileZillaApp::CFileZillaApp()");
 }
 
 CFileZillaApp::~CFileZillaApp()
@@ -75,9 +59,10 @@ CFileZillaApp::~CFileZillaApp()
 }
 
 #ifdef __WXMSW__
+namespace {
 extern "C"
 {
-	typedef HRESULT (WINAPI *t_SetCurrentProcessExplicitAppUserModelID)(PCWSTR AppID);
+	typedef HRESULT(WINAPI *t_SetCurrentProcessExplicitAppUserModelID)(PCWSTR AppID);
 }
 
 static void SetAppId()
@@ -98,6 +83,26 @@ static void SetAppId()
 	pSetCurrentProcessExplicitAppUserModelID(_T("FileZilla.Client.AppID"));
 }
 
+static bool InitWinsock()
+{
+	WSADATA d{};
+	int res = WSAStartup((2 << 8) | 8, &d);
+	if (res != 0) {
+		int err = WSAGetLastError();
+		wxString msg = wxString::Format(_("Could not initialize Winsock (%d): %s"), err, wxSysErrorMsg(err));
+		wxMessageBoxEx(msg, _("Failed to initialize networking"), wxICON_EXCLAMATION);
+		return false;
+	}
+
+	return true;
+}
+
+static void UninitWinsock()
+{
+	WSACleanup();
+}
+}
+
 #endif //__WXMSW__
 
 void CFileZillaApp::InitLocale()
@@ -143,11 +148,11 @@ void CFileZillaApp::InitLocale()
 			for( language = GetFallbackLocale(language); !language.empty(); language = GetFallbackLocale(language) ) {
 				const wxLanguageInfo* fallbackInfo = wxLocale::FindLanguageInfo(language);
 				if( fallbackInfo && SetLocale(fallbackInfo->Language )) {
-					COptions::Get()->SetOption(OPTION_LANGUAGE, language);
+					COptions::Get()->SetOption(OPTION_LANGUAGE, language.ToStdWstring());
 					return;
 				}
 			}
-			COptions::Get()->SetOption(OPTION_LANGUAGE, wxString());
+			COptions::Get()->SetOption(OPTION_LANGUAGE, std::wstring());
 			if (pInfo && !pInfo->Description.empty())
 				wxMessageBoxEx(wxString::Format(_("Failed to set language to %s (%s), using default system language"), pInfo->Description, language), _("Failed to change language"), wxICON_EXCLAMATION);
 			else
@@ -157,15 +162,32 @@ void CFileZillaApp::InitLocale()
 	}
 }
 
+namespace {
+std::wstring translator(char const* const t)
+{
+	return wxGetTranslation(t).ToStdWstring();
+}
+
+std::wstring translator_pf(char const* const singular, char const* const plural, int64_t n)
+{
+	// wxGetTranslation does not support 64bit ints on 32bit systems.
+	if (n < 0) {
+		n = -n;
+	}
+	return wxGetTranslation(singular, plural, (sizeof(unsigned int) < 8 && n > 1000000000) ? (1000000000 + n % 1000000000) : n).ToStdWstring();
+}
+}
+
 bool CFileZillaApp::OnInit()
 {
-	AddStartupProfileRecord(_T("CFileZillaApp::OnInit()"));
+	AddStartupProfileRecord("CFileZillaApp::OnInit()");
 
-	srand( (unsigned)time( NULL ) );
+	fz::set_translators(translator, translator_pf);
 
 #ifdef __WXMSW__
-	// Need to call WSAStartup. Let wx do that for us
-	wxSocketBase::Initialize();
+	if (!InitWinsock()) {
+		return false;
+	}
 
 	SetAppId();
 #endif
@@ -187,8 +209,9 @@ bool CFileZillaApp::OnInit()
 #endif
 
 	int cmdline_result = ProcessCommandLine();
-	if (!cmdline_result)
+	if (!cmdline_result) {
 		return false;
+	}
 
 	LoadLocales();
 
@@ -203,6 +226,10 @@ bool CFileZillaApp::OnInit()
 
 	COptions::Init();
 
+#ifdef USE_MAC_SANDBOX
+	OSXSandboxUserdirs::Get().Load();
+#endif
+
 	InitLocale();
 
 #ifndef _DEBUG
@@ -268,39 +295,17 @@ int CFileZillaApp::OnExit()
 	CSessionManager::Uninit();
 #endif
 #ifdef __WXMSW__
-	wxSocketBase::Shutdown();
+	UninitWinsock();
 #endif
 	return wxApp::OnExit();
 }
 
-bool CFileZillaApp::FileExists(const wxString& file) const
+bool CFileZillaApp::FileExists(std::wstring const& file) const
 {
-	int pos = file.Find('*');
-	if (pos < 0)
-		return wxFileExists(file);
-
-	wxASSERT(pos > 0);
-	wxASSERT(file[pos - 1] == '/');
-	wxASSERT(file.size() > static_cast<size_t>(pos + 1) && file[pos + 1] == '/');
-
-	wxLogNull nullLog;
-	wxDir dir(file.Left(pos));
-	if (!dir.IsOpened())
-		return false;
-
-	wxString subDir;
-	bool found = dir.GetFirst(&subDir, _T(""), wxDIR_DIRS);
-	while (found) {
-		if (FileExists(file.Left(pos) + subDir + file.Mid(pos + 1)))
-			return true;
-
-		found = dir.GetNext(&subDir);
-	}
-
-	return false;
+	return fz::local_filesys::get_file_type(fz::to_native(file), true) == fz::local_filesys::file;
 }
 
-CLocalPath CFileZillaApp::GetDataDir(wxString fileToFind) const
+CLocalPath CFileZillaApp::GetDataDir(std::wstring fileToFind) const
 {
 	/*
 	 * Finding the resources in all cases is a difficult task,
@@ -314,9 +319,10 @@ CLocalPath CFileZillaApp::GetDataDir(wxString fileToFind) const
 	 */
 
 #ifdef __WXMAC__
-	CLocalPath path(wxStandardPaths::Get().GetDataDir());
-	if (FileExists(path.GetPath() + fileToFind))
+	CLocalPath path(wxStandardPaths::Get().GetDataDir().ToStdWstring());
+	if (FileExists(path.GetPath() + fileToFind)) {
 		return path;
+	}
 
 	return CLocalPath();
 #else
@@ -368,27 +374,33 @@ CLocalPath CFileZillaApp::GetDataDir(wxString fileToFind) const
 	// For each path, check for the resources
 	wxPathList::const_iterator node;
 	for (node = pathList.begin(); node != pathList.end(); ++node) {
-		wxString cur = CLocalPath(*node).GetPath();
-		if (FileExists(cur + fileToFind))
+		auto const cur = CLocalPath(node->ToStdWstring()).GetPath();
+		if (FileExists(cur + fileToFind)) {
 			return CLocalPath(cur);
-		if (FileExists(cur + _T("share/filezilla/") + fileToFind))
+		}
+		if (FileExists(cur + _T("share/filezilla/") + fileToFind)) {
 			return CLocalPath(cur + _T("/share/filezilla"));
-		if (FileExists(cur + _T("filezilla/") + fileToFind))
+		}
+		if (FileExists(cur + _T("filezilla/") + fileToFind)) {
 			return CLocalPath(cur + _T("filezilla"));
+		}
 	}
 
 	for (node = pathList.begin(); node != pathList.end(); ++node) {
-		wxString cur = CLocalPath(*node).GetPath();
-		if (FileExists(cur + _T("../") + fileToFind))
+		auto const cur = CLocalPath(node->ToStdWstring()).GetPath();
+		if (FileExists(cur + _T("../") + fileToFind)) {
 			return CLocalPath(cur + _T("/.."));
-		if (FileExists(cur + _T("../share/filezilla/") + fileToFind))
+		}
+		if (FileExists(cur + _T("../share/filezilla/") + fileToFind)) {
 			return CLocalPath(cur + _T("../share/filezilla"));
+		}
 	}
 
 	for (node = pathList.begin(); node != pathList.end(); ++node) {
-		wxString cur = CLocalPath(*node).GetPath();
-		if (FileExists(cur + _T("../../") + fileToFind))
+		auto const cur = CLocalPath(node->ToStdWstring()).GetPath();
+		if (FileExists(cur + _T("../../") + fileToFind)) {
 			return CLocalPath(cur + _T("../.."));
+		}
 	}
 
 	return CLocalPath();
@@ -397,8 +409,8 @@ CLocalPath CFileZillaApp::GetDataDir(wxString fileToFind) const
 
 bool CFileZillaApp::LoadResourceFiles()
 {
-	AddStartupProfileRecord(_T("CFileZillaApp::LoadResourceFiles"));
-	m_resourceDir = GetDataDir(_T("resources/theme.xml"));
+	AddStartupProfileRecord("CFileZillaApp::LoadResourceFiles");
+	m_resourceDir = GetDataDir(_T("resources/defaultfilters.xml"));
 
 	wxImage::AddHandler(new wxPNGHandler());
 
@@ -416,57 +428,28 @@ bool CFileZillaApp::LoadResourceFiles()
 	pResource->SetFlags(pResource->GetFlags() | wxXRC_NO_RELOADING);
 #endif
 
-	pResource->AddHandler(new wxMenuXmlHandler);
-	pResource->AddHandler(new wxMenuBarXmlHandler);
-	pResource->AddHandler(new wxDialogXmlHandler);
-	pResource->AddHandler(new wxPanelXmlHandler);
-	pResource->AddHandler(new wxSizerXmlHandler);
-	pResource->AddHandler(new wxButtonXmlHandler);
-	pResource->AddHandler(new wxBitmapButtonXmlHandler);
-	pResource->AddHandler(new wxStaticTextXmlHandler);
-	pResource->AddHandler(new wxStaticBoxXmlHandler);
-	pResource->AddHandler(new wxStaticBitmapXmlHandler);
-	pResource->AddHandler(new wxTreeCtrlXmlHandler);
-	pResource->AddHandler(new wxListCtrlXmlHandler);
-	pResource->AddHandler(new wxCheckListBoxXmlHandler);
-	pResource->AddHandler(new wxChoiceXmlHandler);
-	pResource->AddHandler(new wxGaugeXmlHandler);
-	pResource->AddHandler(new wxCheckBoxXmlHandler);
-	pResource->AddHandler(new wxSpinCtrlXmlHandler);
-	pResource->AddHandler(new wxRadioButtonXmlHandler);
-	pResource->AddHandler(new wxNotebookXmlHandler);
-	pResource->AddHandler(new wxTextCtrlXmlHandlerEx);
-	pResource->AddHandler(new wxListBoxXmlHandler);
-	pResource->AddHandler(new wxToolBarXmlHandlerEx);
-	pResource->AddHandler(new wxStaticLineXmlHandler);
-	pResource->AddHandler(new wxScrolledWindowXmlHandler);
-	pResource->AddHandler(new wxHyperlinkCtrlXmlHandler);
-	pResource->AddHandler(new wxAnimationCtrlXmlHandler);
-	pResource->AddHandler(new wxStdDialogButtonSizerXmlHandler);
+	InitHandlers(*pResource);
 
-	if (CLocalFileSystem::GetFileType(m_resourceDir.GetPath() + _T("xrc/resources.xrc")) == CLocalFileSystem::file) {
-		pResource->LoadFile(m_resourceDir.GetPath() + _T("xrc/resources.xrc"));
-	}
-	else {
-		CLocalFileSystem fs;
-		wxString dir = m_resourceDir.GetPath() + _T("xrc/");
-		bool found = fs.BeginFindFiles(dir, false);
-		while (found) {
-			wxString name;
-			found = fs.GetNextFile(name);
-			if (name.Right(4) != _T(".xrc")) {
-				continue;
-			}
-			pResource->LoadFile(dir + name);
-		}
-	}
+	fz::local_filesys fs;
+	std::wstring dir = m_resourceDir.GetPath() + L"xrc/";
+	pResource->LoadFile(wxString(dir + L"certificate.xrc"));
+	pResource->LoadFile(wxString(dir + L"dialogs.xrc"));
+	pResource->LoadFile(wxString(dir + L"inputdialog.xrc"));
+	pResource->LoadFile(wxString(dir + L"menus.xrc"));
+	pResource->LoadFile(wxString(dir + L"netconfwizard.xrc"));
+	pResource->LoadFile(wxString(dir + L"settings.xrc"));
+	pResource->LoadFile(wxString(dir + L"sitemanager.xrc"));
+	pResource->LoadFile(wxString(dir + L"update.xrc"));
+
+	// Useful for XRC files with embedded image data.
+	wxFileSystem::AddHandler(new wxFileSystemBlobHandler);
 
 	return true;
 }
 
 bool CFileZillaApp::InitDefaultsDir()
 {
-	AddStartupProfileRecord(_T("InitDefaultsDir"));
+	AddStartupProfileRecord("InitDefaultsDir");
 #ifdef __WXGTK__
 	m_defaultsDir = COptions::GetUnadjustedSettingsDir();
 	if( m_defaultsDir.empty() || !wxFileName::FileExists(m_defaultsDir.GetPath() + _T("fzdefaults.xml"))) {
@@ -488,22 +471,22 @@ bool CFileZillaApp::InitDefaultsDir()
 
 bool CFileZillaApp::LoadLocales()
 {
-	AddStartupProfileRecord(_T("CFileZillaApp::LoadLocales"));
+	AddStartupProfileRecord("CFileZillaApp::LoadLocales");
 #ifndef __WXMAC__
-	m_localesDir = GetDataDir(_T("../locale/*/filezilla.mo"));
+	m_localesDir = GetDataDir(_T("../locale/de/filezilla.mo"));
 	if (m_localesDir.empty())
-		m_localesDir = GetDataDir(_T("../locale/*/LC_MESSAGES/filezilla.mo"));
+		m_localesDir = GetDataDir(_T("../locale/de/LC_MESSAGES/filezilla.mo"));
 	if (!m_localesDir.empty()) {
 		m_localesDir.ChangePath( _T("../locale") );
 	}
 	else {
-		m_localesDir = GetDataDir(_T("locales/*/filezilla.mo"));
+		m_localesDir = GetDataDir(_T("locales/de/filezilla.mo"));
 		if (!m_localesDir.empty()) {
 			m_localesDir.AddSegment(_T("locales"));
 		}
 	}
 #else
-	m_localesDir.SetPath(wxStandardPaths::Get().GetDataDir() + _T("/locales"));
+	m_localesDir.SetPath(wxStandardPaths::Get().GetDataDir().ToStdWstring() + _T("/locales"));
 #endif
 
 	if (!m_localesDir.empty()) {
@@ -521,21 +504,31 @@ bool CFileZillaApp::SetLocale(int language)
 	auto pLocale = std::make_unique<wxLocale>();
 	wxLogNull log;
 	pLocale->Init(language);
-	if (!pLocale->IsOk() || !pLocale->AddCatalog(_T("filezilla"))) {
+	if (!pLocale->IsOk()) {
 		return false;
 	}
 
-	// Now unload old locale
-	// We unload new locale as well, else the internal locale chain in wxWidgets get's broken.
-	pLocale.reset();
-	m_pLocale.reset();
-
-	// Finally load new one
-	pLocale = std::make_unique<wxLocale>();
-	pLocale->Init(language);
-	if (!pLocale->IsOk() || !pLocale->AddCatalog(_T("filezilla"))) {
+	// Load catalog. Only fail if a non-default language has been selected
+	if (!pLocale->AddCatalog(_T("filezilla")) && language != wxLANGUAGE_DEFAULT) {
 		return false;
 	}
+
+	if (m_pLocale) {
+		// Now unload old locale
+		// We unload new locale as well, else the internal locale chain in wxWidgets get's broken.
+		pLocale.reset();
+		m_pLocale.reset();
+
+		// Finally load new one
+		pLocale = std::make_unique<wxLocale>();
+		pLocale->Init(language);
+		if (!pLocale->IsOk()) {
+			return false;
+		}
+		else if (!pLocale->AddCatalog(_T("filezilla")) && language != wxLANGUAGE_DEFAULT) {
+			return false;
+		}
+	}
 	m_pLocale = std::move(pLocale);
 
 	return true;
@@ -581,7 +574,7 @@ CWrapEngine* CFileZillaApp::GetWrapEngine()
 
 void CFileZillaApp::CheckExistsFzsftp()
 {
-	AddStartupProfileRecord(_T("CFileZillaApp::CheckExistsFzstp"));
+	AddStartupProfileRecord("FileZillaApp::CheckExistsFzstp");
 	// Get the correct path to the fzsftp executable
 
 #ifdef __WXMAC__
@@ -590,7 +583,7 @@ void CFileZillaApp::CheckExistsFzsftp()
 	if (pos != -1)
 		executable = executable.Left(pos);
 	executable += _T("/fzsftp");
-	if (!wxFileName::FileExists(executable))
+	if (!wxFileName::FileExists(executable.ToStdWstring()))
 	{
 		wxMessageBoxEx(wxString::Format(_("%s could not be found. Without this component of FileZilla, SFTP will not work.\n\nPlease download FileZilla again. If this problem persists, please submit a bug report."), executable),
 			_("File not found"), wxICON_ERROR);
@@ -609,7 +602,7 @@ void CFileZillaApp::CheckExistsFzsftp()
 	// First check the FZ_FZSFTP environment variable
 	wxString executable;
 	if (wxGetEnv(_T("FZ_FZSFTP"), &executable)) {
-		if (wxFileName::FileExists(executable))
+		if (wxFileName::FileExists(executable.ToStdWstring()))
 			found = true;
 	}
 
@@ -683,7 +676,7 @@ void CFileZillaApp::CheckExistsFzsftp()
 	}
 #endif
 
-	COptions::Get()->SetOption(OPTION_FZSFTP_EXECUTABLE, executable);
+	COptions::Get()->SetOption(OPTION_FZSFTP_EXECUTABLE, executable.ToStdWstring());
 }
 
 #ifdef __WXMSW__
@@ -700,7 +693,7 @@ extern "C" BOOL CALLBACK EnumWindowCallback(HWND hwnd, LPARAM)
 
 int CFileZillaApp::ProcessCommandLine()
 {
-	AddStartupProfileRecord(_T("CFileZillaApp::ProcessCommandLine"));
+	AddStartupProfileRecord("CFileZillaApp::ProcessCommandLine");
 	m_pCommandLine = std::make_unique<CCommandLine>(argc, argv);
 	int res = m_pCommandLine->Parse() ? 1 : -1;
 
@@ -726,36 +719,35 @@ int CFileZillaApp::ProcessCommandLine()
 	return res;
 }
 
-void CFileZillaApp::AddStartupProfileRecord(const wxString& msg)
+void CFileZillaApp::AddStartupProfileRecord(std::string const& msg)
 {
-	if (!m_profilingActive)
+	if (!m_profile_start) {
 		return;
+	}
 
-	m_startupProfile.push_back(std::make_pair(wxDateTime::UNow(), msg));
+	m_startupProfile.emplace_back(fz::monotonic_clock::now(), msg);
 }
 
 void CFileZillaApp::ShowStartupProfile()
 {
-	m_profilingActive = false;
+	if (m_profile_start && m_pCommandLine && m_pCommandLine->HasSwitch(CCommandLine::debug_startup)) {
+		wxString msg = _T("Profile:\n");
+		for (auto const& p : m_startupProfile) {
+			auto const diff = p.first - m_profile_start;
 
-	std::list<std::pair<wxDateTime, wxString> > profile;
-	profile.swap(m_startupProfile);
-
-	if (m_pCommandLine && !m_pCommandLine->HasSwitch(CCommandLine::debug_startup))
-		return;
-
-	wxString msg = _T("Profile:\n");
-	for (auto const& p : profile) {
-		msg += p.first.Format(_T("%Y-%m-%d %H:%M:%S %l"));
-		msg += _T(" ");
-		msg += p.second;
-		msg += _T("\n");
+			msg += std::to_wstring(diff.get_milliseconds());
+			msg += _T(" ");
+			msg += fz::to_wstring(p.second);
+			msg += _T("\n");
+		}
+		wxMessageBoxEx(msg);
 	}
 
-	wxMessageBoxEx(msg);
+	m_profile_start = fz::monotonic_clock();
+	m_startupProfile.clear();
 }
 
-wxString CFileZillaApp::GetSettingsFile(wxString const& name) const
+std::wstring CFileZillaApp::GetSettingsFile(std::wstring const& name) const
 {
 	return COptions::Get()->GetOption(OPTION_DEFAULT_SETTINGSDIR) + name + _T(".xml");
 }
