@@ -1,16 +1,19 @@
 #include <filezilla.h>
-#include "LocalTreeView.h"
-#include "queue.h"
+
+#include "dndobjects.h"
+#include "dragdropmanager.h"
+#include "drop_target_ex.h"
 #include "filezillaapp.h"
 #include "filter.h"
 #include "file_utils.h"
-#include <wx/dnd.h>
-#include "dndobjects.h"
+#include "graphics.h"
 #include "inputdialog.h"
-#include "local_filesys.h"
-#include "dragdropmanager.h"
-#include "drop_target_ex.h"
+#include "LocalTreeView.h"
 #include "Options.h"
+#include "queue.h"
+#include "themeprovider.h"
+
+#include <libfilezilla/local_filesys.hpp>
 
 #ifdef __WXMSW__
 #include <wx/msw/registry.h>
@@ -24,11 +27,11 @@
 class CTreeItemData : public wxTreeItemData
 {
 public:
-	CTreeItemData(const wxString& known_subdir) : m_known_subdir(known_subdir) {}
-	wxString m_known_subdir;
+	CTreeItemData(std::wstring const& known_subdir) : m_known_subdir(known_subdir) {}
+	std::wstring m_known_subdir;
 };
 
-class CLocalTreeViewDropTarget : public CScrollableDropTarget<wxTreeCtrlEx>
+class CLocalTreeViewDropTarget final : public CScrollableDropTarget<wxTreeCtrlEx>
 {
 public:
 	CLocalTreeViewDropTarget(CLocalTreeView* pLocalTreeView)
@@ -57,8 +60,9 @@ public:
 		const wxString& dir = m_pLocalTreeView->GetDirFromItem(item);
 
 #ifdef __WXMSW__
-		if (dir == _T("/"))
+		if (dir == _T("/")) {
 			return wxString();
+		}
 #endif
 
 		return dir;
@@ -69,8 +73,9 @@ public:
 		int flags = 0;
 		wxTreeItemId hit = m_pLocalTreeView->HitTest(point, flags);
 
-		if (flags & (wxTREE_HITTEST_ABOVE | wxTREE_HITTEST_BELOW | wxTREE_HITTEST_NOWHERE | wxTREE_HITTEST_TOLEFT | wxTREE_HITTEST_TORIGHT))
+		if (flags & (wxTREE_HITTEST_ABOVE | wxTREE_HITTEST_BELOW | wxTREE_HITTEST_NOWHERE | wxTREE_HITTEST_TOLEFT | wxTREE_HITTEST_TORIGHT)) {
 			return wxTreeItemId();
+		}
 
 		return hit;
 	}
@@ -89,7 +94,7 @@ public:
 		if (!hit)
 			return wxDragNone;
 
-		const CLocalPath path(GetDirFromItem(hit));
+		const CLocalPath path(GetDirFromItem(hit).ToStdWstring());
 		if (path.empty() || !path.IsWriteable())
 			return wxDragNone;
 
@@ -101,7 +106,7 @@ public:
 			pDragDropManager->pDropTarget = m_pLocalTreeView;
 
 		if (m_pDataObject->GetReceivedFormat() == m_pFileDataObject->GetFormat())
-			m_pLocalTreeView->m_pState->HandleDroppedFiles(m_pFileDataObject, path, def == wxDragCopy);
+			m_pLocalTreeView->m_state.HandleDroppedFiles(m_pFileDataObject, path, def == wxDragCopy);
 		else
 		{
 			if (m_pRemoteDataObject->GetProcessId() != (int)wxGetProcessId())
@@ -110,13 +115,13 @@ public:
 				return wxDragNone;
 			}
 
-			if (!m_pLocalTreeView->m_pState->GetServer() || !m_pRemoteDataObject->GetServer().EqualsNoPass(*m_pLocalTreeView->m_pState->GetServer()))
+			if (!m_pLocalTreeView->m_state.GetServer() || !m_pRemoteDataObject->GetServer().EqualsNoPass(*m_pLocalTreeView->m_state.GetServer()))
 			{
 				wxMessageBoxEx(_("Drag&drop between different servers has not been implemented yet."));
 				return wxDragNone;
 			}
 
-			if (!m_pLocalTreeView->m_pState->DownloadDroppedFiles(m_pRemoteDataObject, path))
+			if (!m_pLocalTreeView->m_state.DownloadDroppedFiles(m_pRemoteDataObject, path))
 				return wxDragNone;
 		}
 
@@ -132,12 +137,14 @@ public:
 		ClearDropHighlight();
 
 		wxTreeItemId hit = GetHit(wxPoint(x, y));
-		if (!hit)
+		if (!hit) {
 			return false;
+		}
 
-		const wxString dir = GetDirFromItem(hit);
-		if (dir.empty() || !CLocalPath(dir).IsWriteable())
+		std::wstring const dir = GetDirFromItem(hit).ToStdWstring();
+		if (dir.empty() || !CLocalPath(dir).IsWriteable()) {
 			return false;
+		}
 
 		return true;
 	}
@@ -235,19 +242,20 @@ EVT_CHAR(CLocalTreeView::OnChar)
 EVT_MENU(XRCID("ID_OPEN"), CLocalTreeView::OnMenuOpen)
 END_EVENT_TABLE()
 
-CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState *pState, CQueueView *pQueueView)
+CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState& state, CQueueView *pQueueView)
 	: wxTreeCtrlEx(parent, id, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxTR_LINES_AT_ROOT | wxTR_HAS_BUTTONS | wxNO_BORDER),
-	CSystemImageList(16),
-	CStateEventHandler(pState),
+	CSystemImageList(CThemeProvider::GetIconSize(iconSizeSmall).x),
+	CStateEventHandler(state),
 	m_pQueueView(pQueueView)
 {
-	wxGetApp().AddStartupProfileRecord(_T("CLocalTreeView::CLocalTreeView"));
+	wxGetApp().AddStartupProfileRecord("CLocalTreeView::CLocalTreeView");
 #ifdef __WXMAC__
 	SetFont(wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT));
 #endif
 
-	pState->RegisterHandler(this, STATECHANGE_LOCAL_DIR);
-	pState->RegisterHandler(this, STATECHANGE_APPLYFILTER);
+	state.RegisterHandler(this, STATECHANGE_LOCAL_DIR);
+	state.RegisterHandler(this, STATECHANGE_APPLYFILTER);
+	state.RegisterHandler(this, STATECHANGE_SERVER);
 
 	SetImageList(GetSystemImageList());
 
@@ -269,6 +277,8 @@ CLocalTreeView::CLocalTreeView(wxWindow* parent, wxWindowID id, CState *pState,
 #endif
 
 	SetDropTarget(new CLocalTreeViewDropTarget(this));
+
+	m_windowTinter = std::make_unique<CWindowTinter>(*this);
 }
 
 CLocalTreeView::~CLocalTreeView()
@@ -397,9 +407,9 @@ wxTreeItemId CLocalTreeView::GetSubdir(wxTreeItemId parent, const wxString& subD
 
 bool CLocalTreeView::DisplayDrives(wxTreeItemId parent)
 {
-	wxGetApp().AddStartupProfileRecord(_T("CLocalTreeView::DisplayDrives"));
+	wxGetApp().AddStartupProfileRecord("CLocalTreeView::DisplayDrives");
 
-	std::list<wxString> drives = CVolumeDescriptionEnumeratorThread::GetDrives();
+	std::list<std::wstring> drives = CVolumeDescriptionEnumeratorThread::GetDrives();
 
 	m_pVolumeEnumeratorThread = new CVolumeDescriptionEnumeratorThread(this);
 	if (m_pVolumeEnumeratorThread->Failed()) {
@@ -407,34 +417,31 @@ bool CLocalTreeView::DisplayDrives(wxTreeItemId parent)
 		m_pVolumeEnumeratorThread = 0;
 	}
 
-	for( std::list<wxString>::const_iterator it = drives.begin(); it != drives.end(); ++it ) {
+	for (auto it = drives.begin(); it != drives.end(); ++it) {
 		wxString drive = *it;
 		if (drive.Right(1) == _T("\\"))
 			drive = drive.RemoveLast();
 
-		wxGetApp().AddStartupProfileRecord(wxString::Format(_T("CLocalTreeView::DisplayDrives adding drive %s"), drive.c_str()));
-		wxTreeItemId item = AppendItem(parent, drive, GetIconIndex(iconType::dir, drive));
+		wxTreeItemId item = AppendItem(parent, drive, GetIconIndex(iconType::dir, _T(""), false));
 		AppendItem(item, _T(""));
-		SortChildren(parent);
 	}
+	SortChildren(parent);
 
-	wxGetApp().AddStartupProfileRecord(_T("CLocalTreeView::DisplayDrives adding drives done"));
+	wxGetApp().AddStartupProfileRecord("CLocalTreeView::DisplayDrives adding drives done");
 
 	return true;
 }
 
 #endif
 
-void CLocalTreeView::DisplayDir(wxTreeItemId parent, const wxString& dirname, const wxString& knownSubdir /*=_T("")*/)
+void CLocalTreeView::DisplayDir(wxTreeItemId parent, const wxString& dirname, std::wstring const& knownSubdir)
 {
-	CLocalFileSystem local_filesystem;
+	fz::local_filesys local_filesys;
 
 	{
 		wxLogNull log;
-		if (!local_filesystem.BeginFindFiles(dirname, true))
-		{
-			if (!knownSubdir.empty())
-			{
+		if (!local_filesys.begin_find_files(fz::to_native(dirname), true)) {
+			if (!knownSubdir.empty()) {
 				wxTreeItemId item = GetSubdir(parent, knownSubdir);
 				if (item != wxTreeItemId())
 					return;
@@ -468,33 +475,36 @@ void CLocalTreeView::DisplayDir(wxTreeItemId parent, const wxString& dirname, co
 
 	bool matchedKnown = false;
 
-	wxString file;
+	fz::native_string file;
 	bool wasLink;
 	int attributes;
 	bool is_dir;
 	static int64_t const size(-1);
-	CDateTime date;
-	while (local_filesystem.GetNextFile(file, wasLink, is_dir, 0, &date, &attributes)) {
+	fz::datetime date;
+	while (local_filesys.get_next_file(file, wasLink, is_dir, 0, &date, &attributes)) {
 		wxASSERT(is_dir);
 		if (file.empty()) {
 			wxGetApp().DisplayEncodingWarning();
 			continue;
 		}
+		std::wstring wfile = fz::to_wstring(file);
 
-		wxString fullName = dirname + file;
+		std::wstring fullName = dirname.ToStdWstring() + wfile;
 #ifdef __WXMSW__
-		if (file.CmpNoCase(knownSubdir))
+		if (fz::stricmp(wfile, knownSubdir))
 #else
-		if (file != knownSubdir)
+		if (wfile != knownSubdir)
 #endif
 		{
-			if (filter.FilenameFiltered(file, dirname, true, size, true, attributes, date))
+			if (filter.FilenameFiltered(wfile, dirname, true, size, true, attributes, date)) {
 				continue;
+			}
 		}
-		else
+		else {
 			matchedKnown = true;
+		}
 
-		wxTreeItemId item = AppendItem(parent, file, GetIconIndex(iconType::dir, fullName),
+		wxTreeItemId item = AppendItem(parent, wfile, GetIconIndex(iconType::dir, fullName),
 #ifdef __WXMSW__
 				-1
 #else
@@ -527,29 +537,30 @@ wxString CLocalTreeView::HasSubdir(const wxString& dirname)
 
 	CFilterManager filter;
 
-	CLocalFileSystem local_filesystem;
-	if (!local_filesystem.BeginFindFiles(dirname, true))
+	fz::local_filesys local_filesys;
+	if (!local_filesys.begin_find_files(fz::to_native(dirname), true)) {
 		return wxString();
+	}
 
-	wxString file;
+	fz::native_string file;
 	bool wasLink;
 	int attributes;
 	bool is_dir;
 	static int64_t const size(-1);
-	CDateTime date;
-	while (local_filesystem.GetNextFile(file, wasLink, is_dir, 0, &date, &attributes))
-	{
+	fz::datetime date;
+	while (local_filesys.get_next_file(file, wasLink, is_dir, 0, &date, &attributes)) {
 		wxASSERT(is_dir);
-		if (file.empty())
-		{
+		if (file.empty()) {
 			wxGetApp().DisplayEncodingWarning();
 			continue;
 		}
 
-		if (filter.FilenameFiltered(file, dirname, true, size, true, attributes, date))
+		std::wstring wfile = fz::to_wstring(file);
+		if (filter.FilenameFiltered(wfile, dirname, true, size, true, attributes, date)) {
 			continue;
+		}
 
-		return file;
+		return wfile;
 	}
 
 	return wxString();
@@ -559,22 +570,19 @@ wxTreeItemId CLocalTreeView::MakeSubdirs(wxTreeItemId parent, wxString dirname,
 {
 	const wxString& separator = wxFileName::GetPathSeparator();
 
-	while (!subDir.empty())
-	{
+	while (!subDir.empty()) {
 		int pos = subDir.Find(separator);
 		wxString segment;
-		if (pos == -1)
-		{
+		if (pos == -1) {
 			segment = subDir;
 			subDir = _T("");
 		}
-		else
-		{
+		else {
 			segment = subDir.Left(pos);
 			subDir = subDir.Mid(pos + 1);
 		}
 
-		DisplayDir(parent, dirname, segment);
+		DisplayDir(parent, dirname, segment.ToStdWstring());
 
 		wxTreeItemId item = GetSubdir(parent, segment);
 		if (!item)
@@ -702,10 +710,10 @@ void CLocalTreeView::RefreshListing()
 		}
 	}
 #else
-	t_dir dir;
-	dir.dir = separator;
-	dir.item = GetRootItem();
-	dirsToCheck.push_back(dir);
+	t_dir root_dir;
+	root_dir.dir = separator;
+	root_dir.item = GetRootItem();
+	dirsToCheck.push_back(root_dir);
 #endif
 
 	CFilterManager filter;
@@ -715,8 +723,8 @@ void CLocalTreeView::RefreshListing()
 		dirsToCheck.pop_front();
 
 		// Step 1: Check if directory exists
-		CLocalFileSystem local_filesystem;
-		if (!local_filesystem.BeginFindFiles(dir.dir, true)) {
+		fz::local_filesys local_filesys;
+		if (!local_filesys.begin_find_files(fz::to_native(dir.dir), true)) {
 			// Dir does exist (listed in parent) but may not be accessible.
 			// Recurse into children anyhow, they might be accessible again.
 			wxTreeItemIdValue value;
@@ -730,24 +738,26 @@ void CLocalTreeView::RefreshListing()
 		}
 
 		// Step 2: Enumerate subdirectories on disk and sort them
-		std::vector<wxString> dirs;
+		std::vector<std::wstring> dirs;
 
-		wxString file;
+		fz::native_string file;
 		static int64_t const size(-1);
 		bool was_link;
 		bool is_dir;
 		int attributes;
-		CDateTime date;
-		while (local_filesystem.GetNextFile(file, was_link, is_dir, 0, &date, &attributes)) {
+		fz::datetime date;
+		while (local_filesys.get_next_file(file, was_link, is_dir, 0, &date, &attributes)) {
 			if (file.empty()) {
 				wxGetApp().DisplayEncodingWarning();
 				continue;
 			}
 
-			if (filter.FilenameFiltered(file, dir.dir, true, size, true, attributes, date))
+			std::wstring wfile = fz::to_wstring(file);
+			if (filter.FilenameFiltered(wfile, dir.dir, true, size, true, attributes, date)) {
 				continue;
+			}
 
-			dirs.push_back(file);
+			dirs.emplace_back(std::move(wfile));
 		}
 		auto const& sortFunc = CFileListCtrlSortBase::GetCmpFunction(m_nameSortMode);
 		std::sort(dirs.begin(), dirs.end(), [&](auto const& lhs, auto const& rhs) { return sortFunc(lhs, rhs) < 0; });
@@ -835,27 +845,34 @@ void CLocalTreeView::OnSelectionChanged(wxTreeEvent& event)
 	}
 
 	wxTreeItemId item = event.GetItem();
-	if (!item)
+	if (!item) {
 		return;
+	}
 
-	wxString dir = GetDirFromItem(item);
+	std::wstring dir = GetDirFromItem(item).ToStdWstring();
 
-	wxString error;
-	if (!m_pState->SetLocalDir(dir, &error)) {
-		if (!error.empty())
+	std::wstring error;
+	if (!m_state.SetLocalDir(dir, &error)) {
+		if (!error.empty()) {
 			wxMessageBoxEx(error, _("Failed to change directory"), wxICON_INFORMATION);
-		else
+		}
+		else {
 			wxBell();
+		}
 		m_setSelection = true;
 		SelectItem(event.GetOldItem());
 		m_setSelection = false;
 	}
 }
 
-void CLocalTreeView::OnStateChange(CState*, enum t_statechange_notifications notification, const wxString&, const void*)
+void CLocalTreeView::OnStateChange(t_statechange_notifications notification, const wxString&, const void*)
 {
-	if (notification == STATECHANGE_LOCAL_DIR)
-		SetDir(m_pState->GetLocalDir().GetPath());
+	if (notification == STATECHANGE_LOCAL_DIR) {
+		SetDir(m_state.GetLocalDir().GetPath());
+	}
+	else if (notification == STATECHANGE_SERVER) {
+		m_windowTinter->SetBackgroundTint(m_state.GetSite().m_colour);
+	}
 	else {
 		wxASSERT(notification == STATECHANGE_APPLYFILTER);
 		RefreshListing();
@@ -864,6 +881,10 @@ void CLocalTreeView::OnStateChange(CState*, enum t_statechange_notifications not
 
 void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 {
+	if (COptions::Get()->GetOptionVal(OPTION_DND_DISABLED) != 0) {
+		return;
+	}
+
 	wxTreeItemId item = event.GetItem();
 	if (!item)
 		return;
@@ -908,7 +929,7 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 	{
 		// We only need to refresh local side if the operation got handled
 		// externally, the internal handlers do this for us already
-		m_pState->RefreshLocal();
+		m_state.RefreshLocal();
 	}
 }
 
@@ -916,35 +937,24 @@ void CLocalTreeView::OnBeginDrag(wxTreeEvent& event)
 
 wxString CLocalTreeView::GetSpecialFolder(int folder, int &iconIndex, int &openIconIndex)
 {
-	LPITEMIDLIST list;
-	if (SHGetSpecialFolderLocation((HWND)GetHandle(), folder, &list) != S_OK)
-		return wxString();
+	wxString name;
 
-	SHFILEINFO shFinfo;
-	if (!SHGetFileInfo((LPCTSTR)list, 0, &shFinfo, sizeof(shFinfo), SHGFI_PIDL | SHGFI_ICON | SHGFI_SMALLICON))
-		return wxString();
+	LPITEMIDLIST list{};
+	if (SHGetSpecialFolderLocation((HWND)GetHandle(), folder, &list) == S_OK) {
+		SHFILEINFO shFinfo{};
+		if (SHGetFileInfo((LPCTSTR)list, 0, &shFinfo, sizeof(shFinfo), SHGFI_PIDL | SHGFI_ICON | SHGFI_SMALLICON) != 0) {
+			DestroyIcon(shFinfo.hIcon);
+			iconIndex = shFinfo.iIcon;
+		}
 
-	DestroyIcon(shFinfo.hIcon);
-	iconIndex = shFinfo.iIcon;
+		if (SHGetFileInfo((LPCTSTR)list, 0, &shFinfo, sizeof(shFinfo), SHGFI_PIDL | SHGFI_ICON | SHGFI_SMALLICON | SHGFI_OPENICON | SHGFI_DISPLAYNAME) != 0) {
+			DestroyIcon(shFinfo.hIcon);
+			openIconIndex = shFinfo.iIcon;
+			name = shFinfo.szDisplayName;
+		}
 
-	if (!SHGetFileInfo((LPCTSTR)list, 0, &shFinfo, sizeof(shFinfo), SHGFI_PIDL | SHGFI_ICON | SHGFI_SMALLICON | SHGFI_OPENICON | SHGFI_DISPLAYNAME))
-		return wxString();
-
-	DestroyIcon(shFinfo.hIcon);
-	openIconIndex = shFinfo.iIcon;
-
-	wxString name = shFinfo.szDisplayName;
-
-	LPMALLOC pMalloc;
-	SHGetMalloc(&pMalloc);
-
-	if (pMalloc)
-	{
-		pMalloc->Free(list);
-		pMalloc->Release();
+		CoTaskMemFree(list);
 	}
-	else
-		wxLogLastError(wxT("SHGetMalloc"));
 
 	return name;
 }
@@ -989,30 +999,33 @@ bool CLocalTreeView::CreateRoot()
 
 void CLocalTreeView::OnVolumesEnumerated(wxCommandEvent& event)
 {
-	if (!m_pVolumeEnumeratorThread)
+	if (!m_pVolumeEnumeratorThread) {
 		return;
+	}
 
 	std::list<CVolumeDescriptionEnumeratorThread::t_VolumeInfo> volumeInfo;
 	volumeInfo = m_pVolumeEnumeratorThread->GetVolumes();
 
-	if (event.GetEventType() == fzEVT_VOLUMESENUMERATED)
-	{
+	if (event.GetEventType() == fzEVT_VOLUMESENUMERATED) {
 		delete m_pVolumeEnumeratorThread;
 		m_pVolumeEnumeratorThread = 0;
 	}
 
-	for (std::list<CVolumeDescriptionEnumeratorThread::t_VolumeInfo>::const_iterator iter = volumeInfo.begin(); iter != volumeInfo.end(); ++iter)
-	{
+	for (std::list<CVolumeDescriptionEnumeratorThread::t_VolumeInfo>::const_iterator iter = volumeInfo.begin(); iter != volumeInfo.end(); ++iter) {
 		wxString drive = iter->volume;
 
 		wxTreeItemIdValue tmp;
 		wxTreeItemId item = GetFirstChild(m_drives, tmp);
-		while (item)
-		{
+		while (item) {
 			wxString name = GetItemText(item);
-			if (name == drive || name.Left(drive.Len() + 1) == drive + _T(" "))
-			{
-				SetItemText(item, drive + _T(" (") + iter->volumeName + _T(")"));
+			if (name == drive || name.Left(drive.Len() + 1) == drive + _T(" ")) {
+
+				if (!iter->volumeName.empty()) {
+					SetItemText(item, drive + _T(" (") + iter->volumeName + _T(")"));
+				}
+				if (iter->icon != -1) {
+					SetItemImage(item, iter->icon);
+				}
 				break;
 			}
 			item = GetNextChild(m_drives, tmp);
@@ -1025,22 +1038,26 @@ void CLocalTreeView::OnVolumesEnumerated(wxCommandEvent& event)
 void CLocalTreeView::OnContextMenu(wxTreeEvent& event)
 {
 	m_contextMenuItem = event.GetItem();
-	if (!m_contextMenuItem.IsOk())
+	if (!m_contextMenuItem.IsOk()) {
 		return;
+	}
 
 	wxMenu* pMenu = wxXmlResource::Get()->LoadMenu(_T("ID_MENU_LOCALTREE"));
-	if (!pMenu)
+	if (!pMenu) {
 		return;
+	}
 
-	const CLocalPath path(GetDirFromItem(m_contextMenuItem));
+	CLocalPath const path(GetDirFromItem(m_contextMenuItem).ToStdWstring());
 
 	const bool hasParent = path.HasParent();
 	const bool writeable = path.IsWriteable();
 
-	const bool remoteConnected = m_pState->IsRemoteConnected() && !m_pState->GetRemotePath().empty();
+	const bool remoteConnected = m_state.IsRemoteConnected() && !m_state.GetRemotePath().empty();
 
-	pMenu->Enable(XRCID("ID_UPLOAD"), hasParent && remoteConnected);
-	pMenu->Enable(XRCID("ID_ADDTOQUEUE"), hasParent && remoteConnected);
+	bool const idle = m_state.GetLocalRecursiveOperation() && !m_state.GetLocalRecursiveOperation()->IsActive();
+
+	pMenu->Enable(XRCID("ID_UPLOAD"), hasParent && remoteConnected && idle);
+	pMenu->Enable(XRCID("ID_ADDTOQUEUE"), hasParent && remoteConnected && idle);
 	pMenu->Enable(XRCID("ID_MKDIR"), writeable);
 	pMenu->Enable(XRCID("ID_DELETE"), writeable && hasParent);
 	pMenu->Enable(XRCID("ID_RENAME"), writeable && hasParent);
@@ -1051,26 +1068,43 @@ void CLocalTreeView::OnContextMenu(wxTreeEvent& event)
 
 void CLocalTreeView::OnMenuUpload(wxCommandEvent& event)
 {
-	if (!m_contextMenuItem.IsOk())
+	auto recursiveOperation = m_state.GetLocalRecursiveOperation();
+	if (recursiveOperation->IsActive()) {
 		return;
+	}
 
-	CLocalPath path(GetDirFromItem(m_contextMenuItem));
-
-	if (!path.HasParent())
+	if (!m_contextMenuItem.IsOk()) {
 		return;
+	}
 
-	if (!m_pState->IsRemoteConnected())
+	CLocalPath path(GetDirFromItem(m_contextMenuItem).ToStdWstring());
+
+	if (!path.HasParent()) {
 		return;
+	}
 
-	const CServer server = *m_pState->GetServer();
-	CServerPath remotePath = m_pState->GetRemotePath();
-	if (remotePath.empty())
+	if (!m_state.IsRemoteConnected()) {
 		return;
+	}
 
-	if (!remotePath.ChangePath(GetItemText(m_contextMenuItem)))
+	const CServer server = *m_state.GetServer();
+	CServerPath remotePath = m_state.GetRemotePath();
+	if (remotePath.empty()) {
 		return;
+	}
 
-	m_pQueueView->QueueFolder(event.GetId() == XRCID("ID_ADDTOQUEUE"), false, path, remotePath, server);
+	if (!remotePath.ChangePath(GetItemText(m_contextMenuItem).ToStdWstring())) {
+		return;
+	}
+
+	local_recursion_root root;
+	root.add_dir_to_visit(path, remotePath);
+	recursiveOperation->AddRecursionRoot(std::move(root));
+
+	bool const queue_only = event.GetId() == XRCID("ID_ADDTOQUEUE");
+
+	CFilterManager filter;
+	recursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_transfer, filter.GetActiveFilters(), !queue_only);
 }
 
 // Create a new Directory
@@ -1079,26 +1113,27 @@ void CLocalTreeView::OnMenuMkdir(wxCommandEvent&)
 	wxString newdir = MenuMkdir();
 	if (!newdir.empty()) {
 		RefreshListing();
-		m_pState->RefreshLocal();
+		m_state.RefreshLocal();
 	}
 }
 
 // Create a new Directory and enter the new Directory
 void CLocalTreeView::OnMenuMkdirChgDir(wxCommandEvent&)
 {
-	wxString newdir = MenuMkdir();
+	std::wstring newdir = MenuMkdir().ToStdWstring();
 	if (newdir.empty()) {
 		return;
 	}
 
 	// OnMenuEnter
-	wxString error;
-	if (!m_pState->SetLocalDir(newdir, &error))
-	{
-		if (!error.empty())
+	std::wstring error;
+	if (!m_state.SetLocalDir(newdir, &error)) {
+		if (!error.empty()) {
 			wxMessageBoxEx(error, _("Failed to change directory"), wxICON_INFORMATION);
-		else
+		}
+		else {
 			wxBell();
+		}
 	}
 }
 
@@ -1106,33 +1141,35 @@ void CLocalTreeView::OnMenuMkdirChgDir(wxCommandEvent&)
 // Returns the name of the new directory
 wxString CLocalTreeView::MenuMkdir()
 {
-	if (!m_contextMenuItem.IsOk())
+	if (!m_contextMenuItem.IsOk()) {
 		return wxString();
+	}
 
-	wxString path = GetDirFromItem(m_contextMenuItem);
-	if (path.Last() != wxFileName::GetPathSeparator())
+	std::wstring path = GetDirFromItem(m_contextMenuItem).ToStdWstring();
+	if (!path.empty() && path.back() != wxFileName::GetPathSeparator()) {
 		path += wxFileName::GetPathSeparator();
+	}
 
-	if (!CLocalPath(path).IsWriteable())
-	{
+	if (!CLocalPath(path).IsWriteable()) {
 		wxBell();
 		return wxString();
 	}
 
 	CInputDialog dlg;
-	if (!dlg.Create(this, _("Create directory"), _("Please enter the name of the directory which should be created:")))
+	if (!dlg.Create(this, _("Create directory"), _("Please enter the name of the directory which should be created:"))) {
 		return wxString();
+	}
 
 	wxString newName = _("New directory");
 	dlg.SetValue(path + newName);
-	dlg.SelectText(path.Len(), path.Len() + newName.Len());
+	dlg.SelectText(path.size(), path.size() + newName.size());
 
-	if (dlg.ShowModal() != wxID_OK)
+	if (dlg.ShowModal() != wxID_OK) {
 		return wxString();
+	}
 
 	wxFileName fn(dlg.GetValue(), _T(""));
-	if (!fn.Normalize(wxPATH_NORM_ALL, path))
-	{
+	if (!fn.Normalize(wxPATH_NORM_ALL, path)) {
 		wxBell();
 		return wxString();
 	}
@@ -1163,7 +1200,7 @@ void CLocalTreeView::OnMenuRename(wxCommandEvent&)
 	}
 #endif
 
-	CLocalPath path(GetDirFromItem(m_contextMenuItem));
+	CLocalPath path(GetDirFromItem(m_contextMenuItem).ToStdWstring());
 	if (!path.HasParent() || !path.IsWriteable()) {
 		wxBell();
 		return;
@@ -1174,39 +1211,47 @@ void CLocalTreeView::OnMenuRename(wxCommandEvent&)
 
 void CLocalTreeView::OnMenuDelete(wxCommandEvent&)
 {
-	if (!m_contextMenuItem.IsOk())
-		return;
-
-	wxString path = GetDirFromItem(m_contextMenuItem);
-
-	CLocalPath local_path(path);
-	if (!local_path.HasParent() || !local_path.IsWriteable())
-		return;
-
-	if (!CLocalFileSystem::RecursiveDelete(path, this))
-		wxGetApp().DisplayEncodingWarning();
-
-	wxTreeItemId item = GetSelection();
-	while (item && item != m_contextMenuItem)
-		item = GetItemParent(item);
-
-	if (!item) {
-		if (GetItemParent(m_contextMenuItem) == GetSelection())
-			m_pState->RefreshLocal();
-		else
-			RefreshListing();
+	if (!m_contextMenuItem.IsOk()) {
 		return;
 	}
 
-	if (!path.empty() && path.Last() == wxFileName::GetPathSeparator())
-		path.RemoveLast();
-	int pos = path.Find(wxFileName::GetPathSeparator(), true);
-	if (pos < 1)
-		path = _T("/");
-	else
-		path = path.Left(pos);
+	std::wstring path = GetDirFromItem(m_contextMenuItem).ToStdWstring();
 
-	m_pState->SetLocalDir(path);
+	CLocalPath local_path(path);
+	if (!local_path.HasParent() || !local_path.IsWriteable()) {
+		return;
+	}
+
+	gui_recursive_remove rmd(this);
+	rmd.remove(fz::to_native(path));
+
+	wxTreeItemId item = GetSelection();
+	while (item && item != m_contextMenuItem) {
+		item = GetItemParent(item);
+	}
+
+	if (!item) {
+		if (GetItemParent(m_contextMenuItem) == GetSelection()) {
+			m_state.RefreshLocal();
+		}
+		else {
+			RefreshListing();
+		}
+		return;
+	}
+
+	if (!path.empty() && path.back() == wxFileName::GetPathSeparator()) {
+		path.pop_back();
+	}
+	size_t pos = path.rfind(wxFileName::GetPathSeparator());
+	if (pos == std::wstring::npos || !pos) {
+		path = _T("/");
+	}
+	else {
+		path = path.substr(0, pos);
+	}
+
+	m_state.SetLocalDir(path);
 	RefreshListing();
 }
 
@@ -1223,7 +1268,7 @@ void CLocalTreeView::OnBeginLabelEdit(wxTreeEvent& event)
 	}
 #endif
 
-	CLocalPath path(GetDirFromItem(item));
+	CLocalPath path(GetDirFromItem(item).ToStdWstring());
 
 	if (!path.HasParent() || !path.IsWriteable())
 	{
@@ -1250,7 +1295,7 @@ void CLocalTreeView::OnEndLabelEdit(wxTreeEvent& event)
 	}
 #endif
 
-	wxString path = GetDirFromItem(item);
+	std::wstring path = GetDirFromItem(item).ToStdWstring();
 
 	CLocalPath local_path(path);
 	if (!local_path.HasParent() || !local_path.IsWriteable()) {
@@ -1259,16 +1304,17 @@ void CLocalTreeView::OnEndLabelEdit(wxTreeEvent& event)
 		return;
 	}
 
-	if (!path.empty() && path.Last() == wxFileName::GetPathSeparator())
-		path.RemoveLast();
+	if (!path.empty() && path.back() == wxFileName::GetPathSeparator()) {
+		path.pop_back();
+	}
 
-	int pos = path.Find(wxFileName::GetPathSeparator(), true);
-	wxASSERT(pos != -1);
+	size_t pos = path.rfind(wxFileName::GetPathSeparator());
+	wxASSERT(pos != std::wstring::npos);
 
-	wxString parent = path.Left(pos + 1);
+	std::wstring parent = path.substr(0, pos + 1);
 
 	const wxString& oldName = GetItemText(item);
-	const wxString& newName = event.GetLabel();
+	std::wstring const newName = event.GetLabel().ToStdWstring();
 	if (newName.empty()) {
 		wxBell();
 		event.Veto();
@@ -1277,8 +1323,9 @@ void CLocalTreeView::OnEndLabelEdit(wxTreeEvent& event)
 
 	wxASSERT(parent + oldName == path);
 
-	if (oldName == newName)
+	if (oldName == newName) {
 		return;
+	}
 
 	if (!RenameFile(this, parent, oldName, newName)) {
 		event.Veto();
@@ -1298,15 +1345,15 @@ void CLocalTreeView::OnEndLabelEdit(wxTreeEvent& event)
 	}
 
 	if (item == currentSel) {
-		m_pState->SetLocalDir(parent + newName);
+		m_state.SetLocalDir(parent + newName);
 		return;
 	}
 
-	wxString sub;
+	std::wstring sub;
 
 	wxTreeItemId tmp = currentSel;
 	while (tmp != GetRootItem() && tmp != item) {
-		sub = wxFileName::GetPathSeparator() + GetItemText(tmp) + sub;
+		sub = (wxFileName::GetPathSeparator() + GetItemText(tmp) + sub).ToStdWstring();
 		tmp = GetItemParent(tmp);
 	}
 
@@ -1316,7 +1363,7 @@ void CLocalTreeView::OnEndLabelEdit(wxTreeEvent& event)
 	}
 
 	// Current selection below renamed item
-	m_pState->SetLocalDir(parent + newName + sub);
+	m_state.SetLocalDir(parent + newName + sub);
 }
 
 void CLocalTreeView::OnChar(wxKeyEvent& event)
@@ -1347,7 +1394,7 @@ bool CLocalTreeView::CheckSubdirStatus(wxTreeItemId& item, const wxString& path)
 		CFilterManager filter;
 
 		const int attributes = S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
-		if (!filter.FilenameFiltered(_T("localhost"), path, true, size, true, attributes, CDateTime()))
+		if (!filter.FilenameFiltered(_T("localhost"), path, true, size, true, attributes, fz::datetime()))
 		{
 			if (!child)
 				AppendItem(item, _T(""));
@@ -1357,20 +1404,23 @@ bool CLocalTreeView::CheckSubdirStatus(wxTreeItemId& item, const wxString& path)
 #endif
 
 	if (child) {
-		if (!GetItemText(child).empty())
+		if (!GetItemText(child).empty()) {
 			return false;
+		}
 
 		CTreeItemData* pData = (CTreeItemData*)GetItemData(child);
 		if (pData) {
 			bool wasLink;
 			int attributes;
-			enum CLocalFileSystem::local_fileType type;
-			CDateTime date;
-			if (!path.empty() && path.Last() == CLocalFileSystem::path_separator)
-				type = CLocalFileSystem::GetFileInfo(path + pData->m_known_subdir, wasLink, 0, &date, &attributes);
-			else
-				type = CLocalFileSystem::GetFileInfo(path + CLocalFileSystem::path_separator + pData->m_known_subdir, wasLink, 0, &date, &attributes);
-			if (type == CLocalFileSystem::dir) {
+			fz::local_filesys::type type;
+			fz::datetime date;
+			if (!path.empty() && path.Last() == fz::local_filesys::path_separator) {
+				type = fz::local_filesys::get_file_info(fz::to_native(path + pData->m_known_subdir), wasLink, 0, &date, &attributes);
+			}
+			else {
+				type = fz::local_filesys::get_file_info(fz::to_native(path + fz::local_filesys::path_separator + pData->m_known_subdir), wasLink, 0, &date, &attributes);
+			}
+			if (type == fz::local_filesys::dir) {
 				CFilterManager filter;
 				if (!filter.FilenameFiltered(pData->m_known_subdir, path, true, size, true, attributes, date))
 					return true;
@@ -1381,7 +1431,7 @@ bool CLocalTreeView::CheckSubdirStatus(wxTreeItemId& item, const wxString& path)
 	wxString sub = HasSubdir(path);
 	if (!sub.empty()) {
 		wxTreeItemId subItem = AppendItem(item, _T(""));
-		SetItemData(subItem, new CTreeItemData(sub));
+		SetItemData(subItem, new CTreeItemData(sub.ToStdWstring()));
 	}
 	else if (child)
 		Delete(child);
@@ -1422,8 +1472,8 @@ void CLocalTreeView::OnDevicechange(WPARAM wParam, LPARAM lParam)
 			mask <<= 1;
 		}
 
-		if (GetSelection() == m_drives && m_pState)
-			m_pState->RefreshLocal();
+		if (GetSelection() == m_drives)
+			m_state.RefreshLocal();
 	}
 }
 
@@ -1433,14 +1483,14 @@ wxTreeItemId CLocalTreeView::AddDrive(wxChar letter)
 	drive += _T(":");
 
 	long drivesToHide = CVolumeDescriptionEnumeratorThread::GetDrivesToHide();
-	if( CVolumeDescriptionEnumeratorThread::IsHidden(drive, drivesToHide) ) {
+	if( CVolumeDescriptionEnumeratorThread::IsHidden(drive.wc_str(), drivesToHide) ) {
 		return wxTreeItemId();
 	}
 
 	// Get the label of the drive
 	wxChar volumeName[501];
 	int oldErrorMode = SetErrorMode(SEM_FAILCRITICALERRORS);
-	BOOL res = GetVolumeInformation(drive + _T("\\"), volumeName, 500, 0, 0, 0, 0, 0);
+	BOOL res = GetVolumeInformation((drive + _T("\\")).wc_str(), volumeName, 500, 0, 0, 0, 0, 0);
 	SetErrorMode(oldErrorMode);
 
 	wxString itemLabel = drive;
