@@ -4,62 +4,38 @@
 #include "dndobjects.h"
 #include "queue.h"
 
-#include <libfilezilla.h>
+#include <libfilezilla_engine.h>
 #include <option_change_event_handler.h>
 
 #include <set>
 #include <wx/progdlg.h>
 
 #include "queue_storage.h"
+#include "local_recursive_operation.h"
+#include "notification.h"
 
-class CFolderProcessingEntry
-{
-public:
-	enum t_type
-	{
-		dir,
-		file
-	};
-	const enum t_type m_type;
 
-	CFolderProcessingEntry(enum t_type type) : m_type(type) {}
-	virtual ~CFolderProcessingEntry() {}
-};
+namespace ActionAfterState {
+enum type {
+	None,
+	ShowNotification,
+	RequestAttention,
+	Close,
+	RunCommand,
+	PlaySound,
 
-class t_newEntry : public CFolderProcessingEntry
-{
-public:
-	t_newEntry()
-		: CFolderProcessingEntry(CFolderProcessingEntry::file)
-	{}
+	// On Windows and OS X, wx can reboot or shutdown the system as well.
+	Reboot,
+	Shutdown,
+	Sleep,
 
-	wxString name;
-	int64_t size{};
-	CDateTime time;
-	int attributes{};
-	bool dir{};
-};
-
-enum ActionAfterState
-{
-	ActionAfterState_Disabled,
-	ActionAfterState_Close,
-	ActionAfterState_Disconnect,
-	ActionAfterState_RunCommand,
-	ActionAfterState_ShowMessage,
-	ActionAfterState_PlaySound
-// On Windows and OS X, wx can reboot or shutdown the system as well.
-#if defined(__WXMSW__) || defined(__WXMAC__)
-	,
-	ActionAfterState_Reboot,
-	ActionAfterState_Shutdown,
-	ActionAfterState_Sleep
-#endif
+	Count
 };
+}
 
 class CStatusLineCtrl;
 class CFileItem;
-struct t_EngineData
+struct t_EngineData final
 {
 	t_EngineData()
 		: pEngine()
@@ -105,32 +81,49 @@ struct t_EngineData
 
 class CMainFrame;
 class CStatusLineCtrl;
-class CFolderProcessingThread;
 class CAsyncRequestQueue;
 class CQueue;
 #if WITH_LIBDBUS
 class CDesktopNotification;
+#elif defined(__WXGTK__) || defined(__WXMSW__)
+class wxNotificationMessage;
 #endif
 
-class CQueueView : public CQueueViewBase, public COptionChangeEventHandler
+class CActionAfterBlocker final
+{
+public:
+	CActionAfterBlocker(CQueueView& queueView)
+		: queueView_(queueView)
+	{
+	}
+
+	~CActionAfterBlocker();
+
+private:
+	friend class CQueueView;
+	CQueueView& queueView_;
+	bool trigger_{};
+};
+
+class CQueueView final : public CQueueViewBase, public COptionChangeEventHandler, private EngineNotificationHandler
 {
-	friend class CFolderProcessingThread;
 	friend class CQueueViewDropTarget;
 	friend class CQueueViewFailed;
+	friend class CActionAfterBlocker;
+
 public:
 	CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsyncRequestQueue* pAsyncRequestQueue);
 	virtual ~CQueueView();
 
 	bool QueueFile(const bool queueOnly, const bool download,
-		const wxString& localFile, const wxString& remoteFile,
+		std::wstring const& localFile, std::wstring const& remoteFile,
 		const CLocalPath& localPath, const CServerPath& remotePath,
-		const CServer& server, int64_t size, enum CEditHandler::fileType edit = CEditHandler::none,
+		const CServer& server, int64_t size, CEditHandler::fileType edit = CEditHandler::none,
 		QueuePriority priority = QueuePriority::normal);
 
 	void QueueFile_Finish(const bool start); // Need to be called after QueueFile
-	bool QueueFiles(const bool queueOnly, const CLocalPath& localPath, const CRemoteDataObject& dataObject);
-	int QueueFiles(const std::list<CFolderProcessingEntry*> &entryList, bool queueOnly, bool download, CServerItem* pServerItem, const CFileExistsNotification::OverwriteAction defaultFileExistsAction);
-	bool QueueFolder(bool queueOnly, bool download, const CLocalPath& localPath, const CServerPath& remotePath, const CServer& server);
+	bool QueueFiles(const bool queueOnly, CLocalPath const& localPath, const CRemoteDataObject& dataObject);
+	bool QueueFiles(const bool queueOnly, CServer const& server, CLocalRecursiveOperation::listing const& listing);
 
 	bool empty() const;
 	int IsActive() const { return m_activeMode; }
@@ -138,7 +131,6 @@ public:
 	bool Quit();
 
 	// This sets the default file exists action for all files currently in queue.
-	// This includes queued folders which are yet to be processed
 	void SetDefaultFileExistsAction(CFileExistsNotification::OverwriteAction action, const TransferDirection direction);
 
 	void UpdateItemSize(CFileItem* pItem, int64_t size);
@@ -157,7 +149,7 @@ public:
 
 	void RenameFileInTransfer(CFileZillaEngine *pEngine, const wxString& newName, bool local);
 
-	static wxString ReplaceInvalidCharacters(const wxString& filename);
+	static std::wstring ReplaceInvalidCharacters(std::wstring const& filename);
 
 	// Get the current download speed as the sum of all active downloads.
 	// Unit is byte/s.
@@ -167,6 +159,8 @@ public:
 	// Unit is byte/s.
 	wxFileOffset GetCurrentUploadSpeed();
 
+	std::shared_ptr<CActionAfterBlocker> GetActionAfterBlocker();
+
 protected:
 
 #ifdef __WXMSW__
@@ -180,10 +174,7 @@ protected:
 
 	// Called from TryStartNextTransfer(), checks
 	// whether it is allowed to start another transfer on that server item
-	bool CanStartTransfer(const CServerItem& server_item, struct t_EngineData *&pEngineData);
-
-	bool ProcessFolderItems(int type = -1);
-	void ProcessUploadFolderItems();
+	bool CanStartTransfer(const CServerItem& server_item, t_EngineData *&pEngineData);
 
 	void ProcessReply(t_EngineData* pEngineData, COperationNotification const& notification);
 	void SendNextCommand(t_EngineData& engineData);
@@ -197,9 +188,7 @@ protected:
 		remove
 	};
 
-	enum ActionAfterState GetActionAfterState() const;
-
-	void ResetEngine(t_EngineData& data, const enum ResetReason reason);
+	void ResetEngine(t_EngineData& data, const ResetReason reason);
 	void DeleteEngines();
 
 	virtual bool RemoveItem(CQueueItem* item, bool destroy, bool updateItemCount = true, bool updateSelections = true, bool forward = true);
@@ -216,10 +205,10 @@ protected:
 	void DisplayQueueSize();
 	void SaveQueue();
 
-	bool IsActionAfter(enum ActionAfterState);
+	bool IsActionAfter(ActionAfterState::type);
 	void ActionAfter(bool warned = false);
 #if defined(__WXMSW__) || defined(__WXMAC__)
-	void ActionAfterWarnUser(ActionAfterState s);
+	void ActionAfterWarnUser(ActionAfterState::type s);
 #endif
 
 	void ProcessNotification(t_EngineData* pEngineData, std::unique_ptr<CNotification> && pNotification);
@@ -230,7 +219,7 @@ protected:
 	void TryRefreshListings();
 	CServer m_last_refresh_server;
 	CServerPath m_last_refresh_path;
-	CMonotonicClock m_last_refresh_listing_time;
+	fz::monotonic_clock m_last_refresh_listing_time;
 
 	// Called from Process Reply.
 	// After a disconnect, check if there's another idle engine that
@@ -245,35 +234,23 @@ protected:
 	std::vector<t_EngineData*> m_engineData;
 	std::list<CStatusLineCtrl*> m_statusLineList;
 
-	/*
-	 * List of queued folders used to populate the queue.
-	 * Index 0 for downloads, index 1 for uploads.
-	 * For each type, only the first one can be active at any given time.
-	 */
-	std::list<CFolderScanItem*> m_queuedFolders[2];
-
-	void RemoveQueuedFolderItem(CFolderScanItem* pFolder);
-
-	CFolderProcessingThread *m_pFolderProcessingThread;
-
 	/*
 	 * Don't update status line positions if m_waitStatusLineUpdate is true.
 	 * This assures we are updating the status line positions only once,
 	 * and not multiple times (for example inside a loop).
 	 */
-	bool m_waitStatusLineUpdate;
+	bool m_waitStatusLineUpdate{};
 
 	// Remember last top item in UpdateStatusLinePositions()
-	int m_lastTopItem;
+	int m_lastTopItem{-1};
 
-	int m_activeCount;
-	int m_activeCountDown;
-	int m_activeCountUp;
-	int m_activeMode; // 0 inactive, 1 only immediate transfers, 2 all
-	int m_quit;
+	int m_activeCount{};
+	int m_activeCountDown{};
+	int m_activeCountUp{};
+	int m_activeMode{}; // 0 inactive, 1 only immediate transfers, 2 all
+	int m_quit{};
 
-	enum ActionAfterState m_actionAfterState;
-	wxString m_actionAfterRunCommand;
+	ActionAfterState::type m_actionAfterState;
 #if defined(__WXMSW__) || defined(__WXMAC__)
 	wxTimer* m_actionAfterTimer{};
 	wxProgressDialog* m_actionAfterWarnDialog{};
@@ -281,8 +258,8 @@ protected:
 	int m_actionAfterTimerId{-1};
 #endif
 
-	int64_t m_totalQueueSize;
-	int m_filesWithUnknownSize;
+	int64_t m_totalQueueSize{};
+	int m_filesWithUnknownSize{};
 
 	CMainFrame* m_pMainFrame;
 	CAsyncRequestQueue* m_pAsyncRequestQueue;
@@ -291,8 +268,6 @@ protected:
 
 	virtual void OnPostScroll();
 
-	wxTimer m_folderscan_item_refresh_timer;
-
 	int GetLineHeight();
 	int m_line_height;
 #ifdef __WXMSW__
@@ -304,7 +279,9 @@ protected:
 	void ReleaseExclusiveEngineLock(CFileZillaEngine* pEngine);
 
 #if WITH_LIBDBUS
-	CDesktopNotification* m_desktop_notification;
+	std::unique_ptr<CDesktopNotification> m_desktop_notification;
+#elif defined(__WXGTK__) || defined(__WXMSW__)
+	std::unique_ptr<wxNotificationMessage> m_desktop_notification;
 #endif
 
 	CQueueStorage m_queue_storage;
@@ -313,10 +290,14 @@ protected:
 	// Unit is byte/s.
 	wxFileOffset GetCurrentSpeed(bool countDownload, bool countUpload);
 
+	virtual void OnEngineEvent(CFileZillaEngine* engine);
+	void DoOnEngineEvent(CFileZillaEngine* engine);
+
+	void OnAskPassword();
+
+	std::weak_ptr<CActionAfterBlocker> m_actionAfterBlocker;
+
 	DECLARE_EVENT_TABLE()
-	void OnEngineEvent(wxFzEvent &event);
-	void OnFolderThreadComplete(wxCommandEvent& event);
-	void OnFolderThreadFiles(wxCommandEvent& event);
 	void OnChar(wxKeyEvent& event);
 
 	// Context menu handlers
@@ -326,7 +307,6 @@ protected:
 	void OnRemoveSelected(wxCommandEvent& event);
 	void OnSetDefaultFileExistsAction(wxCommandEvent& event);
 
-	void OnAskPassword(wxCommandEvent& event);
 	void OnTimer(wxTimerEvent& evnet);
 
 	void OnSetPriority(wxCommandEvent& event);
