@@ -42,7 +42,7 @@ bool CAsyncRequestQueue::ProcessDefaults(CFileZillaEngine *pEngine, std::unique_
 			CFileExistsNotification *pFileExistsNotification = static_cast<CFileExistsNotification *>(pNotification.get());
 
 			// Get the action, go up the hierarchy till one is found
-			enum CFileExistsNotification::OverwriteAction action = pFileExistsNotification->overwriteAction;
+			CFileExistsNotification::OverwriteAction action = pFileExistsNotification->overwriteAction;
 			if (action == CFileExistsNotification::unknown)
 				action = CDefaultFileExistsDlg::GetDefault(pFileExistsNotification->download);
 			if (action == CFileExistsNotification::unknown) {
@@ -50,7 +50,7 @@ bool CAsyncRequestQueue::ProcessDefaults(CFileZillaEngine *pEngine, std::unique_
 				if (option < CFileExistsNotification::unknown || option >= CFileExistsNotification::ACTION_COUNT)
 					action = CFileExistsNotification::unknown;
 				else
-					action = (enum CFileExistsNotification::OverwriteAction)option;
+					action = static_cast<CFileExistsNotification::OverwriteAction>(option);
 			}
 
 			// Ask and rename options require user interaction
@@ -123,8 +123,9 @@ bool CAsyncRequestQueue::AddRequest(CFileZillaEngine *pEngine, std::unique_ptr<C
 
 bool CAsyncRequestQueue::ProcessNextRequest()
 {
-	if (m_requestList.empty())
+	if (m_requestList.empty()) {
 		return true;
+	}
 
 	t_queueEntry &entry = m_requestList.front();
 
@@ -146,23 +147,27 @@ bool CAsyncRequestQueue::ProcessNextRequest()
 		}
 		bool canRemember = notification.GetType() == CInteractiveLoginNotification::keyfile;
 
-		if (CLoginManager::Get().GetPassword(notification.server, true, wxString(), notification.GetChallenge(), canRemember))
+		if (CLoginManager::Get().GetPassword(notification.server, true, std::wstring(), notification.GetChallenge(), canRemember)) {
 			notification.passwordSet = true;
+		}
 		else {
 			// Retry with prompt
 
-			if (!CheckWindowState())
+			if (!CheckWindowState()) {
 				return false;
+			}
 
-			if (CLoginManager::Get().GetPassword(notification.server, false, wxString(), notification.GetChallenge(), canRemember))
+			if (CLoginManager::Get().GetPassword(notification.server, false, std::wstring(), notification.GetChallenge(), canRemember)) {
 				notification.passwordSet = true;
+			}
 		}
 
 		entry.pEngine->SetAsyncRequestReply(std::move(entry.pNotification));
 	}
 	else if (entry.pNotification->GetRequestID() == reqId_hostkey || entry.pNotification->GetRequestID() == reqId_hostkeyChanged) {
-		if (!CheckWindowState())
+		if (!CheckWindowState()) {
 			return false;
+		}
 
 		auto & notification = static_cast<CHostKeyNotification&>(*entry.pNotification.get());
 
@@ -170,14 +175,16 @@ bool CAsyncRequestQueue::ProcessNextRequest()
 			notification.m_trust = true;
 			notification.m_alwaysTrust = false;
 		}
-		else
+		else {
 			CVerifyHostkeyDialog::ShowVerificationDialog(m_pMainFrame, notification);
+		}
 
 		entry.pEngine->SetAsyncRequestReply(std::move(entry.pNotification));
 	}
 	else if (entry.pNotification->GetRequestID() == reqId_certificate) {
-		if (!CheckWindowState())
+		if (!CheckWindowState()) {
 			return false;
+		}
 
 		auto & notification = static_cast<CCertificateNotification&>(*entry.pNotification.get());
 		m_pVerifyCertDlg->ShowVerificationDialog(notification);
@@ -199,7 +206,7 @@ bool CAsyncRequestQueue::ProcessFileExistsNotification(t_queueEntry &entry)
 	auto & notification = static_cast<CFileExistsNotification&>(*entry.pNotification.get());
 
 	// Get the action, go up the hierarchy till one is found
-	enum CFileExistsNotification::OverwriteAction action = notification.overwriteAction;
+	CFileExistsNotification::OverwriteAction action = notification.overwriteAction;
 	if (action == CFileExistsNotification::unknown)
 		action = CDefaultFileExistsDlg::GetDefault(notification.download);
 	if (action == CFileExistsNotification::unknown) {
@@ -207,7 +214,7 @@ bool CAsyncRequestQueue::ProcessFileExistsNotification(t_queueEntry &entry)
 		if (option <= CFileExistsNotification::unknown || option >= CFileExistsNotification::ACTION_COUNT)
 			action = CFileExistsNotification::ask;
 		else
-			action = (enum CFileExistsNotification::OverwriteAction)option;
+			action = static_cast<CFileExistsNotification::OverwriteAction>(option);
 	}
 
 	if (action == CFileExistsNotification::ask) {
@@ -278,11 +285,13 @@ bool CAsyncRequestQueue::ProcessFileExistsNotification(t_queueEntry &entry)
 				return false;
 
 			wxString msg;
-			wxString defaultName;
+			std::wstring defaultName;
 			if (notification.download) {
 				msg.Printf(_("The file %s already exists.\nPlease enter a new name:"), notification.localFile);
-				wxFileName fn = notification.localFile;
-				defaultName = fn.GetFullName();
+				CLocalPath fn(notification.localFile, &defaultName);
+				if (fn.empty() || defaultName.empty()) {
+					defaultName = _("new name");
+				}
 			}
 			else {
 				wxString fullName = notification.remotePath.GetPath() + notification.remoteFile;
