@@ -2,7 +2,7 @@
 #include "commandqueue.h"
 #include "Mainfrm.h"
 #include "state.h"
-#include "recursive_operation.h"
+#include "remote_recursive_operation.h"
 #include "loginmanager.h"
 #include "queue.h"
 #include "RemoteListView.h"
@@ -13,11 +13,11 @@ DEFINE_EVENT_TYPE(fzEVT_GRANTEXCLUSIVEENGINEACCESS)
 
 int CCommandQueue::m_requestIdCounter = 0;
 
-CCommandQueue::CCommandQueue(CFileZillaEngine *pEngine, CMainFrame* pMainFrame, CState* pState)
+CCommandQueue::CCommandQueue(CFileZillaEngine *pEngine, CMainFrame* pMainFrame, CState& state)
+	: m_state(state)
 {
 	m_pEngine = pEngine;
 	m_pMainFrame = pMainFrame;
-	m_pState = pState;
 	m_exclusiveEngineRequest = false;
 	m_exclusiveEngineLock = false;
 	m_requestId = 0;
@@ -36,7 +36,7 @@ bool CCommandQueue::Idle(command_origin origin) const
 	if (origin == any) {
 		return m_CommandList.empty();
 	}
-	
+
 	return std::find_if(m_CommandList.begin(), m_CommandList.end(), [origin](CommandInfo const& c) { return c.origin == origin; }) == m_CommandList.end();
 }
 
@@ -50,7 +50,7 @@ void CCommandQueue::ProcessCommand(CCommand *pCommand, CCommandQueue::command_or
 
 	m_CommandList.emplace_back(origin, std::unique_ptr<CCommand>(pCommand));
 	if (m_CommandList.size() == 1) {
-		m_pState->NotifyHandlers(STATECHANGE_REMOTE_IDLE);
+		m_state.NotifyHandlers(STATECHANGE_REMOTE_IDLE);
 		ProcessNextCommand();
 	}
 }
@@ -75,7 +75,7 @@ void CCommandQueue::ProcessNextCommand()
 		// - Interface cannot obtain listing since not connected
 		// - Yet getting operation successful
 		// To keep things flowing, we need to advance the recursive operation.
-		m_pState->GetRecursiveOperationHandler()->NextOperation();
+		m_state.GetRemoteRecursiveOperation()->NextOperation();
 	}
 
 	while (!m_CommandList.empty()) {
@@ -94,10 +94,11 @@ void CCommandQueue::ProcessNextCommand()
 		if (m_exclusiveEngineRequest)
 			GrantExclusiveEngineRequest();
 		else
-			m_pState->NotifyHandlers(STATECHANGE_REMOTE_IDLE);
+			m_state.NotifyHandlers(STATECHANGE_REMOTE_IDLE);
 
-		if (!m_pState->SuccessfulConnect())
-			m_pState->SetServer(0);
+		if (!m_state.SuccessfulConnect()) {
+			m_state.SetSite(Site());
+		}
 	}
 }
 
@@ -113,7 +114,7 @@ bool CCommandQueue::Cancel()
 
 	if (!m_pEngine)	{
 		m_CommandList.clear();
-		m_pState->NotifyHandlers(STATECHANGE_REMOTE_IDLE);
+		m_state.NotifyHandlers(STATECHANGE_REMOTE_IDLE);
 		return true;
 	}
 
@@ -122,7 +123,7 @@ bool CCommandQueue::Cancel()
 		return false;
 	else {
 		m_CommandList.clear();
-		m_pState->NotifyHandlers(STATECHANGE_REMOTE_IDLE);
+		m_state.NotifyHandlers(STATECHANGE_REMOTE_IDLE);
 		return true;
 	}
 }
@@ -148,7 +149,7 @@ void CCommandQueue::ProcessReply(int nReplyCode, Command commandId)
 			return;
 		}
 		if (nReplyCode & FZ_REPLY_PASSWORDFAILED)
-			CLoginManager::Get().CachedPasswordFailed(*m_pState->GetServer());
+			CLoginManager::Get().CachedPasswordFailed(*m_state.GetServer());
 	}
 
 	if (m_CommandList.empty()) {
@@ -173,7 +174,7 @@ void CCommandQueue::ProcessReply(int nReplyCode, Command commandId)
 
 		if (reconnect) {
 			// Try automatic reconnect
-			const CServer* pServer = m_pState->GetServer();
+			const CServer* pServer = m_state.GetServer();
 			if (pServer) {
 				m_CommandList.emplace_front(normal, std::make_unique<CConnectCommand>(*pServer));
 				ProcessNextCommand();
@@ -187,22 +188,22 @@ void CCommandQueue::ProcessReply(int nReplyCode, Command commandId)
 	auto const& commandInfo = m_CommandList.front();
 
 	if (commandInfo.command->GetId() == Command::list && nReplyCode != FZ_REPLY_OK) {
-		if (nReplyCode & FZ_REPLY_LINKNOTDIR) {
+		if ((nReplyCode & FZ_REPLY_LINKNOTDIR) == FZ_REPLY_LINKNOTDIR) {
 			// Symbolic link does not point to a directory. Either points to file
 			// or is completely invalid
 			CListCommand* pListCommand = static_cast<CListCommand*>(commandInfo.command.get());
 			wxASSERT(pListCommand->GetFlags() & LIST_FLAG_LINK);
 
-			m_pState->LinkIsNotDir(pListCommand->GetPath(), pListCommand->GetSubDir());
+			m_state.LinkIsNotDir(pListCommand->GetPath(), pListCommand->GetSubDir());
 		}
 		else {
 			if (commandInfo.origin == recursiveOperation) {
 				// Let the recursive operation handler know if a LIST command failed,
 				// so that it may issue the next command in recursive operations.
-				m_pState->GetRecursiveOperationHandler()->ListingFailed(nReplyCode);
+				m_state.GetRemoteRecursiveOperation()->ListingFailed(nReplyCode);
 			}
 			else {
-				m_pState->ListingFailed(nReplyCode);
+				m_state.ListingFailed(nReplyCode);
 			}
 		}
 		m_CommandList.pop_front();
@@ -220,10 +221,10 @@ void CCommandQueue::ProcessReply(int nReplyCode, Command commandId)
 
 		// If this was an automatic reconnect during a recursive
 		// operation, stop the recursive operation
-		m_pState->GetRecursiveOperationHandler()->StopRecursiveOperation();
+		m_state.GetRemoteRecursiveOperation()->StopRecursiveOperation();
 	}
 	else if (commandInfo.command->GetId() == Command::connect && nReplyCode == FZ_REPLY_OK) {
-		m_pState->SetSuccessfulConnect();
+		m_state.SetSuccessfulConnect();
 		m_CommandList.pop_front();
 	}
 	else
@@ -248,7 +249,7 @@ void CCommandQueue::RequestExclusiveEngine(bool requestExclusive)
 		}
 		if (m_CommandList.empty())
 		{
-			m_pState->NotifyHandlers(STATECHANGE_REMOTE_IDLE);
+			m_state.NotifyHandlers(STATECHANGE_REMOTE_IDLE);
 			GrantExclusiveEngineRequest();
 			return;
 		}
@@ -256,7 +257,7 @@ void CCommandQueue::RequestExclusiveEngine(bool requestExclusive)
 	if (!requestExclusive)
 		m_exclusiveEngineLock = false;
 	m_exclusiveEngineRequest = requestExclusive;
-	m_pState->NotifyHandlers(STATECHANGE_REMOTE_IDLE);
+	m_state.NotifyHandlers(STATECHANGE_REMOTE_IDLE);
 }
 
 void CCommandQueue::GrantExclusiveEngineRequest()
@@ -304,21 +305,25 @@ void CCommandQueue::ProcessDirectoryListing(CDirectoryListingNotification const&
 	if (!listingNotification.GetPath().empty()) {
 		pListing = std::make_shared<CDirectoryListing>();
 		if (listingNotification.Failed() ||
-			m_pState->m_pEngine->CacheLookup(listingNotification.GetPath(), *pListing) != FZ_REPLY_OK)
+			m_state.m_pEngine->CacheLookup(listingNotification.GetPath(), *pListing) != FZ_REPLY_OK)
 		{
 			pListing = std::make_shared<CDirectoryListing>();
 			pListing->path = listingNotification.GetPath();
 			pListing->m_flags |= CDirectoryListing::listing_failed;
-			pListing->m_firstListTime = CMonotonicClock::now();
+			pListing->m_firstListTime = fz::monotonic_clock::now();
 		}
 	}
 
 	if (listingIsRecursive) {
-		if (!listingNotification.Modified() && m_pState->GetRecursiveOperationHandler()->IsActive()) {
-			m_pState->NotifyHandlers(STATECHANGE_REMOTE_DIR_OTHER, wxString(), &pListing);
+		if (!listingNotification.Modified() && m_state.GetRemoteRecursiveOperation()->IsActive()) {
+			m_state.NotifyHandlers(STATECHANGE_REMOTE_DIR_OTHER, wxString(), &pListing);
 		}
 	}
 	else {
-		m_pState->SetRemoteDir(pListing, listingNotification.Modified());
+		m_state.SetRemoteDir(pListing, listingNotification.Modified());
+	}
+
+	if (pListing && !listingNotification.Failed() && m_state.GetServer()) {
+		CContextManager::Get()->ProcessDirectoryListing(*m_state.GetServer(), pListing, listingIsRecursive ? 0 : &m_state);
 	}
 }
