@@ -3,12 +3,13 @@
 #include "context_control.h"
 #include "filelist_statusbar.h"
 #include "filezillaapp.h"
+#include "local_recursive_operation.h"
 #include "LocalListView.h"
 #include "LocalTreeView.h"
 #include "Mainfrm.h"
 #include "Options.h"
 #include "queue.h"
-#include "recursive_operation.h"
+#include "remote_recursive_operation.h"
 #include "recursive_operation_status.h"
 #include "RemoteListView.h"
 #include "RemoteTreeView.h"
@@ -31,8 +32,7 @@ EVT_MENU(XRCID("ID_TABCONTEXT_NEW"), CContextControl::OnTabContextNew)
 END_EVENT_TABLE()
 
 CContextControl::CContextControl(CMainFrame& mainFrame)
-	: CStateEventHandler(0)
-	, m_mainFrame(mainFrame)
+	: m_mainFrame(mainFrame)
 {
 	wxASSERT(!CContextManager::Get()->HandlerCount(STATECHANGE_CHANGEDCONTEXT));
 	CContextManager::Get()->RegisterHandler(this, STATECHANGE_CHANGEDCONTEXT, false);
@@ -50,56 +50,81 @@ void CContextControl::Create(wxWindow *parent)
 
 void CContextControl::CreateTab()
 {
-	wxGetApp().AddStartupProfileRecord(_T("CContextControl::CreateTab"));
-#ifdef __WXMSW__
-	// Some reparenting is being done when creating tabs. Reparenting of frozen windows isn't working
-	// on OS X.
-	wxWindowUpdateLocker lock(this);
-#endif
+	wxGetApp().AddStartupProfileRecord("CContextControl::CreateTab");
 
-	CState* pState = 0;
+	{
+	#ifdef __WXMSW__
+		// Some reparenting is being done when creating tabs. Reparenting of frozen windows isn't working
+		// on OS X.
+		wxWindowUpdateLocker lock(this);
+	#endif
 
-	// See if we can reuse an existing context
-	for (size_t i = 0; i < m_context_controls.size(); i++) {
-		if (m_context_controls[i].tab_index != -1)
-			continue;
+		CState* pState = 0;
 
-		if (m_context_controls[i].pState->IsRemoteConnected() ||
-			!m_context_controls[i].pState->IsRemoteIdle())
-			continue;
+		// See if we can reuse an existing context
+		for (size_t i = 0; i < m_context_controls.size(); i++) {
+			if (m_context_controls[i].used()) {
+				continue;
+			}
 
-		pState = m_context_controls[i].pState;
-		m_context_controls.erase(m_context_controls.begin() + i);
-		if (m_current_context_controls > (int)i)
-			m_current_context_controls--;
-		break;
-	}
-	if (!pState) {
-		pState = CContextManager::Get()->CreateState(m_mainFrame);
-		if (!pState->CreateEngine()) {
-			wxMessageBoxEx(_("Failed to initialize FTP engine"));
+			if (m_context_controls[i].pState->IsRemoteConnected() ||
+				!m_context_controls[i].pState->IsRemoteIdle())
+			{
+				continue;
+			}
+
+			pState = m_context_controls[i].pState;
+			m_context_controls.erase(m_context_controls.begin() + i);
+			if (m_current_context_controls > (int)i) {
+				--m_current_context_controls;
+			}
+			break;
 		}
-	}
-
-	// Restore last server and path
-	CServer last_server;
-	CServerPath last_path;
-	if (COptions::Get()->GetLastServer(last_server) && last_path.SetSafePath(COptions::Get()->GetOption(OPTION_LASTSERVERPATH)))
-		pState->SetLastServer(last_server, last_path);
-
-	CreateContextControls(pState);
-
-	pState->GetRecursiveOperationHandler()->SetQueue(m_mainFrame.GetQueue());
-
-	wxString const localDir = COptions::Get()->GetOption(OPTION_LASTLOCALDIR);
-	if (!pState->SetLocalDir(localDir)) {
-		wxString const homeDir = wxGetHomeDir();
-		if (!pState->SetLocalDir(homeDir)) {
-			pState->SetLocalDir(_T("/"));
+		if (!pState) {
+			pState = CContextManager::Get()->CreateState(m_mainFrame);
+			if (!pState->CreateEngine()) {
+				wxMessageBoxEx(_("Failed to initialize FTP engine"));
+			}
 		}
-	}
 
-	CContextManager::Get()->SetCurrentContext(pState);
+		// Restore last server and path
+		auto * current = GetCurrentControls();
+		if (current) {
+			pState->SetLastSite(current->pState->GetLastSite(), current->pState->GetLastServerPath());
+		}
+		else {
+			CServer last_server;
+			CServerPath last_path;
+			if (COptions::Get()->GetLastServer(last_server) && last_path.SetSafePath(COptions::Get()->GetOption(OPTION_LASTSERVERPATH))) {
+				std::wstring last_site_path = COptions::Get()->GetOption(OPTION_LAST_CONNECTED_SITE);
+
+				std::unique_ptr<Site> site;
+				if (!last_site_path.empty()) {
+					site = CSiteManager::GetSiteByPath(last_site_path, false).first;
+				}
+				if (!site || site->m_server != last_server) {
+					site.reset(new Site);
+					site->m_server = last_server;
+				}
+				pState->SetLastSite(*site, last_path);
+			}
+		}
+
+		CreateContextControls(*pState);
+
+		pState->GetLocalRecursiveOperation()->SetQueue(m_mainFrame.GetQueue());
+		pState->GetRemoteRecursiveOperation()->SetQueue(m_mainFrame.GetQueue());
+
+		std::wstring const localDir = COptions::Get()->GetOption(OPTION_LASTLOCALDIR);
+		if (!pState->SetLocalDir(localDir)) {
+			std::wstring const homeDir = wxGetHomeDir().ToStdWstring();
+			if (!pState->SetLocalDir(homeDir)) {
+				pState->SetLocalDir(_T("/"));
+			}
+		}
+
+		CContextManager::Get()->SetCurrentContext(pState);
+	}
 
 	if (!m_mainFrame.RestoreSplitterPositions())
 		m_mainFrame.SetDefaultSplitterPositions();
@@ -108,9 +133,9 @@ void CContextControl::CreateTab()
 		m_tabs->SetSelection(m_tabs->GetPageCount() - 1);
 }
 
-void CContextControl::CreateContextControls(CState* pState)
+void CContextControl::CreateContextControls(CState& state)
 {
-	wxGetApp().AddStartupProfileRecord(_T("CContextControl::CreateContextControls"));
+	wxGetApp().AddStartupProfileRecord("CContextControl::CreateContextControls");
 	wxWindow* parent = this;
 
 #ifdef __WXGTK__
@@ -125,7 +150,7 @@ void CContextControl::CreateContextControls(CState* pState)
 			m_tabs = new wxAuiNotebookEx();
 
 			wxSize splitter_size = m_context_controls[m_current_context_controls].pViewSplitter->GetSize();
-			m_tabs->Create(this, wxID_ANY, initial_position, splitter_size, wxNO_BORDER | wxAUI_NB_SCROLL_BUTTONS | wxAUI_NB_WINDOWLIST_BUTTON | wxAUI_NB_CLOSE_ON_ALL_TABS);
+			m_tabs->Create(this, wxID_ANY, initial_position, splitter_size, wxNO_BORDER | wxAUI_NB_SCROLL_BUTTONS | wxAUI_NB_WINDOWLIST_BUTTON | wxAUI_NB_CLOSE_ON_ALL_TABS | wxAUI_NB_TAB_MOVE);
 			m_tabs->SetExArtProvider();
 			m_tabs->SetSelectedFont(*wxNORMAL_FONT);
 			m_tabs->SetMeasuringFont(*wxNORMAL_FONT);
@@ -133,6 +158,7 @@ void CContextControl::CreateContextControls(CState* pState)
 			m_context_controls[m_current_context_controls].pViewSplitter->Reparent(m_tabs);
 
 			m_tabs->AddPage(m_context_controls[m_current_context_controls].pViewSplitter, m_context_controls[m_current_context_controls].pState->GetTitle());
+			m_tabs->SetTabColour(0, m_context_controls[m_current_context_controls].pState->GetSite().m_colour);
 			ReplaceWindow(m_context_controls[m_current_context_controls].pViewSplitter, m_tabs);
 
 			m_tabs->Connect(wxEVT_COMMAND_AUINOTEBOOK_PAGE_CHANGED, wxAuiNotebookEventHandler(CContextControl::OnTabChanged), 0, this);
@@ -154,9 +180,9 @@ void CContextControl::CreateContextControls(CState* pState)
 		parent = m_tabs;
 	}
 
-	struct CContextControl::_context_controls context_controls;
+	CContextControl::_context_controls context_controls;
 
-	context_controls.pState = pState;
+	context_controls.pState = &state;
 	context_controls.pViewSplitter = new CSplitterWindowEx(parent, -1, initial_position, wxDefaultSize, wxSP_NOBORDER  | wxSP_LIVE_UPDATE);
 	context_controls.pViewSplitter->SetMinimumPaneSize(50, 100);
 	context_controls.pViewSplitter->SetSashGravity(0.5);
@@ -169,15 +195,15 @@ void CContextControl::CreateContextControls(CState* pState)
 
 	context_controls.pLocalTreeViewPanel = new CView(context_controls.pLocalSplitter);
 	context_controls.pLocalListViewPanel = new CView(context_controls.pLocalSplitter);
-	context_controls.pLocalTreeView = new CLocalTreeView(context_controls.pLocalTreeViewPanel, -1, pState, m_mainFrame.GetQueue());
-	context_controls.pLocalListView = new CLocalListView(context_controls.pLocalListViewPanel, pState, m_mainFrame.GetQueue());
+	context_controls.pLocalTreeView = new CLocalTreeView(context_controls.pLocalTreeViewPanel, -1, state, m_mainFrame.GetQueue());
+	context_controls.pLocalListView = new CLocalListView(context_controls.pLocalListViewPanel, state, m_mainFrame.GetQueue());
 	context_controls.pLocalTreeViewPanel->SetWindow(context_controls.pLocalTreeView);
 	context_controls.pLocalListViewPanel->SetWindow(context_controls.pLocalListView);
 
 	context_controls.pRemoteTreeViewPanel = new CView(context_controls.pRemoteSplitter);
 	context_controls.pRemoteListViewPanel = new CView(context_controls.pRemoteSplitter);
-	context_controls.pRemoteTreeView = new CRemoteTreeView(context_controls.pRemoteTreeViewPanel, -1, pState, m_mainFrame.GetQueue());
-	context_controls.pRemoteListView = new CRemoteListView(context_controls.pRemoteListViewPanel, pState, m_mainFrame.GetQueue());
+	context_controls.pRemoteTreeView = new CRemoteTreeView(context_controls.pRemoteTreeViewPanel, -1, state, m_mainFrame.GetQueue());
+	context_controls.pRemoteListView = new CRemoteListView(context_controls.pRemoteListViewPanel, state, m_mainFrame.GetQueue());
 	context_controls.pRemoteTreeViewPanel->SetWindow(context_controls.pRemoteTreeView);
 	context_controls.pRemoteListViewPanel->SetWindow(context_controls.pRemoteListView);
 
@@ -196,8 +222,11 @@ void CContextControl::CreateContextControls(CState* pState)
 	context_controls.pRemoteListViewPanel->SetStatusBar(pRemoteFilelistStatusBar);
 	context_controls.pRemoteListView->SetFilelistStatusBar(pRemoteFilelistStatusBar);
 
-	auto recursiveStatus = new CRecursiveOperationStatus(context_controls.pRemoteListViewPanel, context_controls.pState);
-	context_controls.pRemoteListViewPanel->SetFooter(recursiveStatus);
+	auto localRecursiveStatus = new CRecursiveOperationStatus(context_controls.pLocalListViewPanel, state, true);
+	context_controls.pLocalListViewPanel->SetFooter(localRecursiveStatus);
+
+	auto remoteRecursiveStatus = new CRecursiveOperationStatus(context_controls.pRemoteListViewPanel, state, false);
+	context_controls.pRemoteListViewPanel->SetFooter(remoteRecursiveStatus);
 
 	const int layout = COptions::Get()->GetOptionVal(OPTION_FILEPANE_LAYOUT);
 	const int swap = COptions::Get()->GetOptionVal(OPTION_FILEPANE_SWAP);
@@ -216,7 +245,7 @@ void CContextControl::CreateContextControls(CState* pState)
 	}
 
 	if (COptions::Get()->GetOptionVal(OPTION_SHOW_TREE_LOCAL)) {
-		context_controls.pLocalViewHeader = new CLocalViewHeader(context_controls.pLocalTreeViewPanel, pState);
+		context_controls.pLocalViewHeader = new CLocalViewHeader(context_controls.pLocalTreeViewPanel, state);
 		context_controls.pLocalTreeViewPanel->SetHeader(context_controls.pLocalViewHeader);
 		if (layout == 3 && swap)
 			context_controls.pLocalSplitter->SplitVertically(context_controls.pLocalListViewPanel, context_controls.pLocalTreeViewPanel);
@@ -227,13 +256,13 @@ void CContextControl::CreateContextControls(CState* pState)
 	}
 	else {
 		context_controls.pLocalTreeViewPanel->Hide();
-		context_controls.pLocalViewHeader = new CLocalViewHeader(context_controls.pLocalListViewPanel, pState);
+		context_controls.pLocalViewHeader = new CLocalViewHeader(context_controls.pLocalListViewPanel, state);
 		context_controls.pLocalListViewPanel->SetHeader(context_controls.pLocalViewHeader);
 		context_controls.pLocalSplitter->Initialize(context_controls.pLocalListViewPanel);
 	}
 
 	if (COptions::Get()->GetOptionVal(OPTION_SHOW_TREE_REMOTE)) {
-		context_controls.pRemoteViewHeader = new CRemoteViewHeader(context_controls.pRemoteTreeViewPanel, pState);
+		context_controls.pRemoteViewHeader = new CRemoteViewHeader(context_controls.pRemoteTreeViewPanel, state);
 		context_controls.pRemoteTreeViewPanel->SetHeader(context_controls.pRemoteViewHeader);
 		if (layout == 3 && !swap)
 			context_controls.pRemoteSplitter->SplitVertically(context_controls.pRemoteListViewPanel, context_controls.pRemoteTreeViewPanel);
@@ -244,7 +273,7 @@ void CContextControl::CreateContextControls(CState* pState)
 	}
 	else {
 		context_controls.pRemoteTreeViewPanel->Hide();
-		context_controls.pRemoteViewHeader = new CRemoteViewHeader(context_controls.pRemoteListViewPanel, pState);
+		context_controls.pRemoteViewHeader = new CRemoteViewHeader(context_controls.pRemoteListViewPanel, state);
 		context_controls.pRemoteListViewPanel->SetHeader(context_controls.pRemoteViewHeader);
 		context_controls.pRemoteSplitter->Initialize(context_controls.pRemoteListViewPanel);
 	}
@@ -263,27 +292,12 @@ void CContextControl::CreateContextControls(CState* pState)
 	m_mainFrame.ConnectNavigationHandler(context_controls.pLocalViewHeader);
 	m_mainFrame.ConnectNavigationHandler(context_controls.pRemoteViewHeader);
 
-	pState->GetComparisonManager()->SetListings(context_controls.pLocalListView, context_controls.pRemoteListView);
+	state.GetComparisonManager()->SetListings(context_controls.pLocalListView, context_controls.pRemoteListView);
 
 	if (m_tabs) {
-		context_controls.tab_index = m_tabs->GetPageCount();
-		m_tabs->AddPage(context_controls.pViewSplitter, pState->GetTitle());
-
-		// Copy reconnect and bookmark information
-		pState->SetLastServer(
-			m_context_controls[m_current_context_controls].pState->GetLastServer(),
-			m_context_controls[m_current_context_controls].pState->GetLastServerPath());
-
-		context_controls.site_bookmarks = m_context_controls[m_current_context_controls].site_bookmarks;
+		m_tabs->AddPage(context_controls.pViewSplitter, state.GetTitle());
 	}
 	else {
-		context_controls.tab_index = 0;
-		context_controls.site_bookmarks = std::make_shared<CContextControl::_context_controls::_site_bookmarks>();
-
-		context_controls.site_bookmarks->path = COptions::Get()->GetOption(OPTION_LAST_CONNECTED_SITE);
-		CSiteManager::GetBookmarks(context_controls.site_bookmarks->path,
-			context_controls.site_bookmarks->bookmarks);
-
 		Initialize(context_controls.pViewSplitter);
 	}
 
@@ -292,22 +306,18 @@ void CContextControl::CreateContextControls(CState* pState)
 
 void CContextControl::OnTabRefresh(wxCommandEvent&)
 {
-	if (m_right_clicked_tab == -1)
+	if (m_right_clicked_tab == -1) {
 		return;
+	}
 
-	for (size_t j = 0; j < m_context_controls.size(); j++)
-	{
-		if (m_context_controls[j].tab_index != m_right_clicked_tab)
-			continue;
-
-		m_context_controls[j].pState->RefreshLocal();
-		m_context_controls[j].pState->RefreshRemote();
-
-		break;
+	auto * controls = GetControlsFromTabIndex(m_right_clicked_tab);
+	if (controls) {
+		controls->pState->RefreshLocal();
+		controls->pState->RefreshRemote();
 	}
 }
 
-struct CContextControl::_context_controls* CContextControl::GetCurrentControls()
+ CContextControl::_context_controls* CContextControl::GetCurrentControls()
 {
 	if (m_current_context_controls == -1)
 		return 0;
@@ -315,11 +325,10 @@ struct CContextControl::_context_controls* CContextControl::GetCurrentControls()
 	return &m_context_controls[m_current_context_controls];
 }
 
-struct CContextControl::_context_controls* CContextControl::GetControlsFromState(CState* pState)
+CContextControl::_context_controls* CContextControl::GetControlsFromState(CState* pState)
 {
 	size_t i = 0;
-	for (i = 0; i < m_context_controls.size(); i++)
-	{
+	for (i = 0; i < m_context_controls.size(); ++i) {
 		if (m_context_controls[i].pState == pState)
 			return &m_context_controls[i];
 	}
@@ -328,24 +337,19 @@ struct CContextControl::_context_controls* CContextControl::GetControlsFromState
 
 bool CContextControl::CloseTab(int tab)
 {
-	if (!m_tabs)
+	if (!m_tabs) {
 		return false;
-	if (tab < 0)
-		return false;
-
-	size_t i = 0;
-	for (i = 0; i < m_context_controls.size(); i++)
-	{
-		if (m_context_controls[i].tab_index == tab)
-			break;
 	}
-	if (i == m_context_controls.size())
+	if (tab < 0 || static_cast<size_t>(tab) >= m_tabs->GetPageCount()) {
 		return false;
+	}
 
-	CState* pState = m_context_controls[i].pState;
 
-	if (!pState->m_pCommandQueue->Idle())
-	{
+	auto *const removeControls = GetControlsFromTabIndex(tab);
+
+	CState *const pState = removeControls->pState;
+
+	if (!pState->m_pCommandQueue->Idle()) {
 		if (wxMessageBoxEx(_("Cannot close tab while busy.\nCancel current operation and close tab?"), _T("FileZilla"), wxYES_NO | wxICON_QUESTION) != wxYES)
 			return false;
 	}
@@ -357,65 +361,47 @@ bool CContextControl::CloseTab(int tab)
 #endif
 
 	pState->m_pCommandQueue->Cancel();
-	pState->GetRecursiveOperationHandler()->StopRecursiveOperation();
+	pState->GetLocalRecursiveOperation()->StopRecursiveOperation();
+	pState->GetRemoteRecursiveOperation()->StopRecursiveOperation();
 
 	pState->GetComparisonManager()->SetListings(0, 0);
 
-	if (m_tabs->GetPageCount() == 2)
-	{
+	if (m_tabs->GetPageCount() == 2) {
 		// Get rid again of tab bar
 		m_tabs->Disconnect(wxEVT_COMMAND_AUINOTEBOOK_PAGE_CHANGED, wxAuiNotebookEventHandler(CContextControl::OnTabChanged), 0, this);
 
 		int keep = tab ? 0 : 1;
+
+		auto * keptControls = GetControlsFromTabIndex(keep);
 		m_tabs->RemovePage(keep);
 
-		size_t j;
-		for (j = 0; j < m_context_controls.size(); j++)
-		{
-			if (m_context_controls[j].tab_index != keep)
-				continue;
+		CContextManager::Get()->SetCurrentContext(keptControls->pState);
 
-			break;
-		}
-
-		m_context_controls[j].pViewSplitter->Reparent(this);
-		ReplaceWindow(m_tabs, m_context_controls[j].pViewSplitter);
-		m_context_controls[j].pViewSplitter->Show();
-		m_context_controls[j].tab_index = 0;
+		keptControls->pViewSplitter->Reparent(this);
+		ReplaceWindow(m_tabs, keptControls->pViewSplitter);
+		keptControls->pViewSplitter->Show();
 
 		wxAuiNotebookEx *tabs = m_tabs;
 		m_tabs = 0;
 
-		m_context_controls[i].tab_index = -1;
-		m_context_controls[i].site_bookmarks.reset();
+		removeControls->pViewSplitter = 0;
 
-		CContextManager::Get()->SetCurrentContext(m_context_controls[j].pState);
+		CContextManager::Get()->SetCurrentContext(keptControls->pState);
 
 		tabs->Destroy();
 	}
-	else
-	{
-		if (pState == CContextManager::Get()->GetCurrentContext())
-		{
+	else {
+		if (pState == CContextManager::Get()->GetCurrentContext()) {
 			int newsel = tab + 1;
-			if (newsel >= (int)m_tabs->GetPageCount())
+			if (newsel >= (int)m_tabs->GetPageCount()) {
 				newsel = m_tabs->GetPageCount() - 2;
-
-			for (size_t j = 0; j < m_context_controls.size(); j++)
-			{
-				if (m_context_controls[j].tab_index != newsel)
-					continue;
-				m_tabs->SetSelection(newsel);
-				CContextManager::Get()->SetCurrentContext(m_context_controls[j].pState);
 			}
+
+			m_tabs->SetSelection(newsel);
+			CContextManager::Get()->SetCurrentContext(GetControlsFromTabIndex(newsel)->pState);
 		}
-		for (size_t j = 0; j < m_context_controls.size(); j++)
-		{
-			if (m_context_controls[j].tab_index > tab)
-				m_context_controls[j].tab_index--;
-		}
-		m_context_controls[i].tab_index = -1;
-		m_context_controls[i].site_bookmarks.reset();
+
+		removeControls->pViewSplitter = 0;
 		m_tabs->DeletePage(tab);
 	}
 
@@ -467,14 +453,13 @@ void CContextControl::OnTabContextCloseOthers(wxCommandEvent&)
 void CContextControl::OnTabClosing_Deferred(wxCommandEvent& event)
 {
 	int tab = event.GetId();
-	if (tab < 0)
-	{
+	if (tab < 0) {
 		tab++;
 		int count = GetTabCount();
-		for (int i = count - 1; i >= 0; i--)
-		{
-			if (i != -tab)
+		for (int i = count - 1; i >= 0; --i) {
+			if (i != -tab) {
 				CloseTab(i);
+			}
 		}
 	}
 	else
@@ -485,17 +470,12 @@ void CContextControl::OnTabClosing_Deferred(wxCommandEvent& event)
 void CContextControl::OnTabChanged(wxAuiNotebookEvent&)
 {
 	int i = m_tabs->GetSelection();
-	if (i < 0 || i >= (int)m_context_controls.size())
+	auto * const controls = GetControlsFromTabIndex(i);
+	if (!controls) {
 		return;
-
-	for (size_t j = 0; j < m_context_controls.size(); j++)
-	{
-		if (m_context_controls[j].tab_index != i)
-			continue;
-
-		CContextManager::Get()->SetCurrentContext(m_context_controls[j].pState);
-		break;
 	}
+
+	CContextManager::Get()->SetCurrentContext(controls->pState);
 }
 
 void CContextControl::OnTabClosing(wxAuiNotebookEvent& event)
@@ -517,12 +497,26 @@ int CContextControl::GetTabCount() const
 	return m_tabs ? m_tabs->GetPageCount() : (m_context_controls.empty() ? 0 : 1);
 }
 
-struct CContextControl::_context_controls* CContextControl::GetControlsFromTabIndex(int i)
+CContextControl::_context_controls* CContextControl::GetControlsFromTabIndex(int i)
 {
-	for (size_t j = 0; j < m_context_controls.size(); j++)
-	{
-		if (m_context_controls[j].tab_index == i)
-			return &m_context_controls[j];
+	if (!m_tabs) {
+		if (i == 0 && !m_context_controls.empty()) {
+			for (auto & controls : m_context_controls) {
+				if (controls.pViewSplitter != 0) {
+					return &controls;
+				}
+			}
+		}
+		return 0;
+	}
+
+	wxWindow* page = m_tabs->GetPage(i);
+	if (page) {
+		for (auto & controls : m_context_controls) {
+			if (controls.pViewSplitter == page) {
+				return &controls;
+			}
+		}
 	}
 
 	return 0;
@@ -530,19 +524,21 @@ struct CContextControl::_context_controls* CContextControl::GetControlsFromTabIn
 
 bool CContextControl::SelectTab(int i)
 {
-	if (i < 0)
+	if (i < 0) {
 		return false;
+	}
 
-	if (!m_tabs)
-	{
-		if (i != 0)
+	if (!m_tabs) {
+		if (i != 0) {
 			return false;
+		}
 
 		return true;
 	}
 
-	if ((int)m_tabs->GetPageCount() <= i)
+	if ((int)m_tabs->GetPageCount() <= i) {
 		return false;
+	}
 
 	m_tabs->SetSelection(i);
 
@@ -551,39 +547,42 @@ bool CContextControl::SelectTab(int i)
 
 void CContextControl::AdvanceTab(bool forward)
 {
-	if (!m_tabs)
+	if (!m_tabs) {
 		return;
+	}
 
 	m_tabs->AdvanceTab(forward);
 }
 
-void CContextControl::OnStateChange(CState* pState, enum t_statechange_notifications notification, const wxString&, const void*)
+void CContextControl::OnStateChange(CState* pState, t_statechange_notifications notification, const wxString&, const void*)
 {
-	if (notification == STATECHANGE_CHANGEDCONTEXT)
-	{
-		if (!pState)
-		{
+	if (notification == STATECHANGE_CHANGEDCONTEXT) {
+		if (!pState) {
 			m_current_context_controls = 0;
 			return;
 		}
 
 		// Get current controls for new current context
-		for (m_current_context_controls = 0; m_current_context_controls < (int)m_context_controls.size(); m_current_context_controls++)
-		{
+		for (m_current_context_controls = 0; m_current_context_controls < (int)m_context_controls.size(); ++m_current_context_controls) {
 			if (m_context_controls[m_current_context_controls].pState == pState)
 				break;
 		}
 		if (m_current_context_controls == (int)m_context_controls.size())
 			m_current_context_controls = -1;
 	}
-	else if (notification == STATECHANGE_SERVER)
-	{
-		if (!m_tabs)
+	else if (notification == STATECHANGE_SERVER) {
+		if (!m_tabs) {
 			return;
+		}
 
 		CContextControl::_context_controls* controls = GetControlsFromState(pState);
-		if (controls && controls->tab_index != -1)
-			m_tabs->SetPageText(controls->tab_index, controls->pState->GetTitle());
+		if (controls && controls->used()) {
+			int i = m_tabs->GetPageIndex(controls->pViewSplitter);
+			if (i != wxNOT_FOUND) {
+				m_tabs->SetTabColour(i, controls->pState->GetSite().m_colour);
+				m_tabs->SetPageText(i, controls->pState->GetTitle());
+			}
+		}
 	}
 }
 
