@@ -3,7 +3,7 @@
 
 #include "dialogex.h"
 
-class CFileExistsDlg: public wxDialogEx
+class CFileExistsDlg final : public wxDialogEx
 {
 	DECLARE_EVENT_TABLE()
 
@@ -14,14 +14,14 @@ public:
 	/// Creation
 	bool Create(wxWindow* parent);
 
-	enum CFileExistsNotification::OverwriteAction GetAction() const;
+	CFileExistsNotification::OverwriteAction GetAction() const;
 	bool Always(bool &directionOnly, bool &queueOnly) const;
 
 protected:
 	/// Creates the controls and sizers
 	bool CreateControls();
 
-	void DisplayFile(bool left, wxString name, int64_t size, CDateTime const& time, wxString const& iconFile);
+	void DisplayFile(bool left, wxString name, int64_t size, fz::datetime const& time, wxString const& iconFile);
 
 	void OnOK(wxCommandEvent& event);
 	void OnCancel(wxCommandEvent& event);
@@ -31,7 +31,7 @@ protected:
 	wxString GetPathEllipsis(wxString path, wxWindow *window);
 
 	CFileExistsNotification *m_pNotification;
-	enum CFileExistsNotification::OverwriteAction m_action;
+	CFileExistsNotification::OverwriteAction m_action;
 	bool m_always{};
 	bool m_directionOnly{};
 	bool m_queueOnly{};
