@@ -130,7 +130,7 @@ void wxListCtrlEx::ScrollTopItem(int item)
 	else if (item >= GetItemCount()) {
 		item = GetItemCount() - 1;
 	}
-	
+
 	const int current = GetTopItem();
 
 	int delta = item - current;
@@ -151,12 +151,12 @@ void wxListCtrlEx::HandlePrefixSearch(wxChar character)
 	wxASSERT(character);
 
 	// Keyboard navigation within items
-	wxDateTime now = wxDateTime::UNow();
-	if (m_prefixSearch_lastKeyPress.IsValid())
-	{
-		wxTimeSpan span = now - m_prefixSearch_lastKeyPress;
-		if (span.GetSeconds() >= 1)
+	fz::datetime now = fz::datetime::now();
+	if (!m_prefixSearch_lastKeyPress.empty()) {
+		fz::duration span = now - m_prefixSearch_lastKeyPress;
+		if (span.get_seconds() >= 1) {
 			m_prefixSearch_prefix = _T("");
+		}
 	}
 	m_prefixSearch_lastKeyPress = now;
 
@@ -174,8 +174,7 @@ void wxListCtrlEx::HandlePrefixSearch(wxChar character)
 	}
 
 	bool beep = false;
-	if (item != -1)
-	{
+	if (item != -1) {
 		wxString text = GetItemText(item, 0);
 		if (text.Length() >= m_prefixSearch_prefix.Length() && !m_prefixSearch_prefix.CmpNoCase(text.Left(m_prefixSearch_prefix.Length())))
 			beep = true;
@@ -496,32 +495,32 @@ void wxListCtrlEx::LoadColumnSettings(int widthsOptionId, int visibilityOptionId
 
 void wxListCtrlEx::SaveColumnSettings(int widthsOptionId, int visibilityOptionId, int sortOptionId)
 {
-	if (widthsOptionId != -1)
+	if (widthsOptionId != -1) {
 		SaveColumnWidths(widthsOptionId);
-
-	if (visibilityOptionId != -1)
-	{
-		wxString visibleColumns;
-		for (unsigned int i = 0; i < m_columnInfo.size(); i++)
-		{
-			if (m_columnInfo[i].shown)
-				visibleColumns += _T("1");
-			else
-				visibleColumns += _T("0");
-		}
-		COptions::Get()->SetOption(visibilityOptionId, visibleColumns);
 	}
 
-	if (sortOptionId != -1)
-	{
+	if (visibilityOptionId != -1) {
+		wxString visibleColumns;
+		for (unsigned int i = 0; i < m_columnInfo.size(); ++i) {
+			if (m_columnInfo[i].shown) {
+				visibleColumns += _T("1");
+			}
+			else {
+				visibleColumns += _T("0");
+			}
+		}
+		COptions::Get()->SetOption(visibilityOptionId, visibleColumns.ToStdWstring());
+	}
+
+	if (sortOptionId != -1) {
 		wxString order;
-		for (unsigned int i = 0; i < m_columnInfo.size(); i++)
-		{
-			if (i)
+		for (unsigned int i = 0; i < m_columnInfo.size(); ++i) {
+			if (i) {
 				order += _T(",");
+			}
 			order += wxString::Format(_T("%d"), m_columnInfo[i].order);
 		}
-		COptions::Get()->SetOption(sortOptionId, order);
+		COptions::Get()->SetOption(sortOptionId, order.ToStdWstring());
 	}
 }
 
@@ -536,30 +535,23 @@ bool wxListCtrlEx::ReadColumnWidths(unsigned int optionId)
 		return true;
 	}
 
-	const unsigned int count = m_columnInfo.size();
+	size_t const columnCount = m_columnInfo.size();
 
 	wxString savedWidths = COptions::Get()->GetOption(optionId);
 	wxStringTokenizer tokens(savedWidths, _T(" "));
-	if (tokens.CountTokens() < count)
-		return false;
 
-	unsigned long* newWidths = new unsigned long[count];
-	for (unsigned int i = 0; i < count; i++)
-	{
-		wxString token = tokens.GetNextToken();
-		if (!token.ToULong(newWidths + i) || newWidths[i] > 5000)
-		{
-			delete [] newWidths;
-			return false;
+	size_t const tokenCount = tokens.CountTokens();
+
+	for (size_t i = 0; i < columnCount; ++i) {
+		unsigned long newWidth{};
+		if (i < tokenCount) {
+			wxString token = tokens.GetNextToken();
+			if (token.ToULong(&newWidth) && newWidth < 10000) {
+				m_columnInfo[i].width = newWidth;
+			}
 		}
-		else if (newWidths[i] < MIN_COLUMN_WIDTH)
-			newWidths[i] = MIN_COLUMN_WIDTH;
 	}
 
-	for (unsigned int i = 0; i < count; i++)
-		m_columnInfo[i].width = newWidths[i];
-
-	delete [] newWidths;
 	return true;
 }
 
@@ -568,13 +560,11 @@ void wxListCtrlEx::SaveColumnWidths(unsigned int optionId)
 	const unsigned int count = m_columnInfo.size();
 
 	wxString widths;
-	for (unsigned int i = 0; i < count; i++)
-	{
+	for (unsigned int i = 0; i < count; ++i) {
 		int width = 0;
 
 		bool found = false;
-		for (int j = 0; j < GetColumnCount(); j++)
-		{
+		for (int j = 0; j < GetColumnCount(); ++j) {
 			if (m_pVisibleColumnMapping[j] != i)
 				continue;
 
@@ -587,11 +577,11 @@ void wxListCtrlEx::SaveColumnWidths(unsigned int optionId)
 	}
 	widths.RemoveLast();
 
-	COptions::Get()->SetOption(optionId, widths);
+	COptions::Get()->SetOption(optionId, widths.ToStdWstring());
 }
 
 
-void wxListCtrlEx::AddColumn(const wxString& name, int align, int initialWidth, bool fixed /*=false*/)
+void wxListCtrlEx::AddColumn(const wxString& name, int align, int initialWidth, bool fixed)
 {
 	wxASSERT(!GetColumnCount());
 
@@ -667,10 +657,8 @@ void wxListCtrlEx::MoveColumn(unsigned int col, unsigned int before)
 void wxListCtrlEx::CreateVisibleColumnMapping()
 {
 	int pos = 0;
-	for (unsigned int j = 0; j < m_columnInfo.size(); j++)
-	{
-		for (unsigned int i = 0; i < m_columnInfo.size(); i++)
-		{
+	for (unsigned int j = 0; j < m_columnInfo.size(); ++j) {
+		for (unsigned int i = 0; i < m_columnInfo.size(); ++i) {
 			const t_columnInfo &column = m_columnInfo[i];
 
 			if (!column.shown)
@@ -692,7 +680,7 @@ public:
 	DECLARE_EVENT_TABLE()
 
 protected:
-	void OnUp(wxCommandEvent& event)
+	void OnUp(wxCommandEvent&)
 	{
 		wxCheckListBox* pListBox = XRCCTRL(*this, "ID_ACTIVE", wxCheckListBox);
 		int sel = pListBox->GetSelection();
@@ -715,7 +703,7 @@ protected:
 		OnSelChanged(evt);
 	}
 
-	void OnDown(wxCommandEvent& event)
+	void OnDown(wxCommandEvent&)
 	{
 		wxCheckListBox* pListBox = XRCCTRL(*this, "ID_ACTIVE", wxCheckListBox);
 		int sel = pListBox->GetSelection();
@@ -740,7 +728,7 @@ protected:
 		OnSelChanged(evt);
 	}
 
-	void OnSelChanged(wxCommandEvent& event)
+	void OnSelChanged(wxCommandEvent&)
 	{
 		wxCheckListBox* pListBox = XRCCTRL(*this, "ID_ACTIVE", wxCheckListBox);
 		int sel = pListBox->GetSelection();
@@ -822,13 +810,14 @@ void wxListCtrlEx::ShowColumnEditor()
 
 int wxListCtrlEx::GetColumnVisibleIndex(int col)
 {
-	if (!m_pVisibleColumnMapping)
+	if (!m_pVisibleColumnMapping) {
 		return -1;
+	}
 
-	for (int i = 0; i < GetColumnCount(); i++)
-	{
-		if (m_pVisibleColumnMapping[i] == (unsigned int)col)
+	for (int i = 0; i < GetColumnCount(); ++i) {
+		if (m_pVisibleColumnMapping[i] == (unsigned int)col) {
 			return i;
+		}
 	}
 
 	return -1;
@@ -836,8 +825,9 @@ int wxListCtrlEx::GetColumnVisibleIndex(int col)
 
 int wxListCtrlEx::GetHeaderSortIconIndex(int col)
 {
-	if (col < 0 || col >= GetColumnCount())
+	if (col < 0 || col >= GetColumnCount()) {
 		return -1;
+	}
 
 #ifdef __WXMSW__
 	HWND hWnd = (HWND)GetHandle();
@@ -847,14 +837,16 @@ int wxListCtrlEx::GetHeaderSortIconIndex(int col)
 	item.mask = HDI_IMAGE | HDI_FORMAT;
 	SendMessage(header, HDM_GETITEM, col, (LPARAM)&item);
 
-	if (!(item.fmt & HDF_IMAGE))
+	if (!(item.fmt & HDF_IMAGE)) {
 		return -1;
+	}
 
 	return item.iImage;
 #else
 	wxListItem item;
-	if (!GetColumn(col, item))
+	if (!GetColumn(col, item)) {
 		return -1;
+	}
 
 	return item.GetImage();
 #endif
@@ -862,113 +854,66 @@ int wxListCtrlEx::GetHeaderSortIconIndex(int col)
 
 void wxListCtrlEx::InitHeaderSortImageList()
 {
-#ifdef __WXMSW__
-	if (wxPlatformInfo::Get().GetOSMajorVersion() >= 6)
-		return;
-
-	// Initialize imagelist for list header
-	m_pHeaderImageList = new wxImageListEx(8, 8, true, 3);
-
-	wxBitmap bmp;
-
-	bmp.LoadFile(wxGetApp().GetResourceDir().GetPath() + _T("up.png"), wxBITMAP_TYPE_PNG);
-	m_pHeaderImageList->Add(bmp);
-	bmp.LoadFile(wxGetApp().GetResourceDir().GetPath() + _T("down.png"), wxBITMAP_TYPE_PNG);
-	m_pHeaderImageList->Add(bmp);
-
-	HWND hWnd = (HWND)GetHandle();
-	if (!hWnd)
-	{
-		delete m_pHeaderImageList;
-		m_pHeaderImageList = 0;
-		return;
-	}
-
-	HWND header = (HWND)SendMessage(hWnd, LVM_GETHEADER, 0, 0);
-	if (!header)
-	{
-		delete m_pHeaderImageList;
-		m_pHeaderImageList = 0;
-		return;
-	}
-
-	TCHAR buffer[1000] = {0};
-	HDITEM item;
-	item.mask = HDI_TEXT;
-	item.pszText = buffer;
-	item.cchTextMax = 999;
-	SendMessage(header, HDM_GETITEM, 0, (LPARAM)&item);
-
-	SendMessage(header, HDM_SETIMAGELIST, 0, (LPARAM)m_pHeaderImageList->GetHandle());
-
-	m_header_icon_index.up = 0;
-	m_header_icon_index.down = 1;
-#else
-
+#ifndef __WXMSW__
 	wxColour colour = wxSystemSettings::GetColour(wxSYS_COLOUR_3DFACE);
 
 	wxString lightness;
-	if (colour.Red() + colour.Green() + colour.Blue() > 3 * 128)
+	if (colour.Red() + colour.Green() + colour.Blue() > 3 * 128) {
 		lightness = _T("DARK");
-	else
+	}
+	else {
 		lightness = _T("LIGHT");
+	}
 
-	wxBitmap bmp;
+	auto * imageList = GetSystemImageList();
+	if (imageList) {
+		wxBitmap bmp;
 
-	bmp = wxArtProvider::GetBitmap(_T("ART_SORT_UP_") + lightness, wxART_OTHER, CThemeProvider::GetIconSize(iconSizeSmall));
-	m_header_icon_index.up = GetSystemImageList()->Add(bmp);
-	bmp = wxArtProvider::GetBitmap(_T("ART_SORT_DOWN_") + lightness, wxART_OTHER, CThemeProvider::GetIconSize(iconSizeSmall));
-	m_header_icon_index.down = GetSystemImageList()->Add(bmp);
+		bmp = wxArtProvider::GetBitmap(_T("ART_SORT_UP_") + lightness, wxART_OTHER, CThemeProvider::GetIconSize(iconSizeSmall));
+		m_header_icon_index.up = imageList->Add(bmp);
+		bmp = wxArtProvider::GetBitmap(_T("ART_SORT_DOWN_") + lightness, wxART_OTHER, CThemeProvider::GetIconSize(iconSizeSmall));
+		m_header_icon_index.down = imageList->Add(bmp);
+	}
 #endif
 }
 
 void wxListCtrlEx::SetHeaderSortIconIndex(int col, int icon)
 {
-	if (col < 0 || col >= GetColumnCount())
+	if (col < 0 || col >= GetColumnCount()) {
 		return;
+	}
 
 #ifdef __WXMSW__
 	HWND hWnd = (HWND)GetHandle();
 	HWND header = (HWND)SendMessage(hWnd, LVM_GETHEADER, 0, 0);
 
-	wxChar buffer[100];
-	HDITEM item = {0};
-	item.mask = HDI_TEXT | HDI_FORMAT;
-	item.pszText = buffer;
-	item.cchTextMax = 99;
+	HDITEM item = {};
+	item.mask = HDI_FORMAT;
 	SendMessage(header, HDM_GETITEM, col, (LPARAM)&item);
-	if (icon != -1)
-	{
-		if (wxPlatformInfo::Get().GetOSMajorVersion() >= 6)
-		{
-			item.fmt &= ~(HDF_IMAGE | HDF_BITMAP_ON_RIGHT | HDF_SORTUP | HDF_SORTDOWN);
-			item.iImage = -1;
-			if (icon)
-				item.fmt |= HDF_SORTDOWN;
-			else
-				item.fmt |= HDF_SORTUP;
+	if (icon != -1) {
+		item.fmt &= ~(HDF_IMAGE | HDF_BITMAP_ON_RIGHT | HDF_SORTUP | HDF_SORTDOWN);
+		item.iImage = -1;
+		if (icon) {
+			item.fmt |= HDF_SORTDOWN;
 		}
-		else
-		{
-			item.fmt &= ~(HDF_SORTUP | HDF_SORTDOWN);
-			item.fmt |= HDF_IMAGE | HDF_BITMAP_ON_RIGHT;
-			item.iImage = icon ? m_header_icon_index.down : m_header_icon_index.up;
-			item.mask |= HDI_IMAGE;
+		else {
+			item.fmt |= HDF_SORTUP;
 		}
 	}
-	else
-	{
+	else {
 		item.fmt &= ~(HDF_IMAGE | HDF_BITMAP_ON_RIGHT | HDF_SORTUP | HDF_SORTDOWN);
 		item.iImage = -1;
 	}
 	SendMessage(header, HDM_SETITEM, col, (LPARAM)&item);
 #else
 	wxListItem item;
-	if (!GetColumn(col, item))
+	if (!GetColumn(col, item)) {
 		return;
+	}
 
-	if( icon != -1 )
+	if (icon != -1) {
 		icon = icon ? m_header_icon_index.down : m_header_icon_index.up;
+	}
 
 	item.SetImage(icon);
 	SetColumn(col, item);
@@ -1057,12 +1002,12 @@ void wxListCtrlEx::OnEndLabelEdit(wxListEvent& event)
 		event.Veto();
 }
 
-bool wxListCtrlEx::OnBeginRename(const wxListEvent& event)
+bool wxListCtrlEx::OnBeginRename(const wxListEvent&)
 {
 	return false;
 }
 
-bool wxListCtrlEx::OnAcceptRename(const wxListEvent& event)
+bool wxListCtrlEx::OnAcceptRename(const wxListEvent&)
 {
 	return false;
 }
@@ -1125,21 +1070,25 @@ bool wxListCtrlEx::MSWOnNotify(int idCtrl, WXLPARAM lParam, WXLPARAM *result)
 		break;
 	case HDN_ITEMCHANGINGA:
 	case HDN_ITEMCHANGINGW:
-		if (m_columnDragging)
-		{
-			if (nmHDR->pitem->mask & HDI_WIDTH && nmHDR->pitem->cxy < MIN_COLUMN_WIDTH)
-			{
+		if (m_columnDragging) {
+			if (nmHDR->pitem->mask & HDI_WIDTH && nmHDR->pitem->cxy < MIN_COLUMN_WIDTH) {
 				*result = 1;
 				return true;
 			}
-			else
-			{
+			else {
 				*result = 0;
 				return false;
 			}
 		}
 		else
 			return false;
+	case HDN_DIVIDERDBLCLICK:
+		{
+			auto event = new wxListEvent(wxEVT_LIST_COL_END_DRAG, GetId());
+			event->SetEventObject(this);
+			QueueEvent(event);
+		}
+		break;
 	}
 
 	return wxListCtrl::MSWOnNotify(idCtrl, lParam, result);
