@@ -2,6 +2,7 @@
 #define __SEARCH_H__
 
 #include "filter_conditions_dialog.h"
+#include "local_recursive_operation.h"
 #include "state.h"
 #include <set>
 
@@ -9,11 +10,18 @@ class CWindowStateManager;
 class CSearchDialogFileList;
 class CQueueView;
 class CFilelistStatusBar;
-class CSearchDialog : protected CFilterConditionsDialog, public CStateEventHandler
+class CSearchDialog final : protected CFilterConditionsDialog, public CStateEventHandler
 {
 	friend class CSearchDialogFileList;
 public:
-	CSearchDialog(wxWindow* parent, CState* pState, CQueueView* pQueue);
+	enum class search_mode
+	{
+		none,
+		local,
+		remote
+	};
+
+	CSearchDialog(wxWindow* parent, CState& state, CQueueView* pQueue);
 	virtual ~CSearchDialog();
 
 	bool Load();
@@ -21,6 +29,7 @@ public:
 
 protected:
 	void ProcessDirectoryListing(std::shared_ptr<CDirectoryListing> const& listing);
+	void ProcessDirectoryListing(CLocalRecursiveOperation::listing const& listing);
 
 	void SetCtrlState();
 
@@ -31,31 +40,32 @@ protected:
 	CSearchDialogFileList *m_results{};
 	CQueueView* m_pQueue;
 
-	virtual void OnStateChange(CState* pState, enum t_statechange_notifications notification, const wxString& data, const void* data2);
+	virtual void OnStateChange(t_statechange_notifications notification, const wxString& data, const void* data2);
 
 	CWindowStateManager* m_pWindowStateManager{};
 
 	CFilter m_search_filter;
 
-	bool m_searching{};
+	search_mode m_searching{};
 
 	CServerPath m_original_dir;
-	CLocalPath m_local_target;
-
-	void ProcessSelection(std::list<int> &selected_files, std::list<CServerPath> &selected_dirs);
 
 	DECLARE_EVENT_TABLE()
 	void OnSearch(wxCommandEvent& event);
 	void OnStop(wxCommandEvent& event);
 	void OnContextMenu(wxContextMenuEvent& event);
 	void OnDownload(wxCommandEvent&);
+	void OnUpload(wxCommandEvent&);
 	void OnEdit(wxCommandEvent&);
 	void OnDelete(wxCommandEvent&);
 	void OnCharHook(wxKeyEvent& event);
+	void OnChangeSearchMode(wxCommandEvent&);
+	void OnGetUrl(wxCommandEvent& event);
 
 	std::set<CServerPath> m_visited;
 
-	CServerPath m_search_root;
+	CLocalPath m_local_search_root;
+	CServerPath m_remote_search_root;
 };
 
 #endif //__SEARCH_H__
